from collections import defaultdict
from collections.abc import Callable, Iterable
from typing import override

from greyhorse.maybe import Just, Maybe, Nothing

from .abc.collectors import Collector, MutCollector, MutNamedCollector, NamedCollector
from .abc.selectors import ListSelector, NamedListSelector


class DictRegistry[K, T](Collector[K, T], ListSelector[K, T]):
    __slots__ = ('_storage',)

    def __init__(self) -> None:
        self._storage: dict[K, T] = {}

    def __len__(self) -> int:
        return len(self._storage)

    @override
    def add(self, key: K, instance: T) -> bool:
        if key in self._storage:
            return instance is self._storage[key]
        self._storage[key] = instance
        return True

    @override
    def has(self, key: K) -> bool:
        return key in self._storage

    @override
    def get(self, key: K) -> Maybe[T]:
        if key not in self._storage:
            return Nothing

        return Just(self._storage[key])

    @override
    def items(self, filter_fn: Callable[[K], bool] | None = None) -> Iterable[tuple[K, T]]:
        for k, v in self._storage.items():
            if not filter_fn or filter_fn(k):
                yield k, v

    def clear(self) -> None:
        while key := next(iter(self._storage), None):
            del self._storage[key]


class MutDictRegistry[K, T](MutCollector[K, T], DictRegistry[K, T]):
    @override
    def remove(self, key: K, instance: T | None = None) -> bool:
        if key not in self._storage:
            return False

        value = self._storage[key]

        if instance is None:
            del self._storage[key]

        else:
            if instance != value:
                return False
            del self._storage[key]

        return True


class ScopedDictRegistry[K, T](DictRegistry[K, T]):
    def __init__(
        self, factory: Callable[[], DictRegistry[K, T]], scope_func: Callable[[], str]
    ) -> None:
        super().__init__()
        self._scope_func = scope_func
        self._storage: dict[str, DictRegistry[K, T]] = defaultdict(factory)

    def _get_registry(self) -> DictRegistry[K, T]:
        key = self._scope_func()
        return self._storage[key]

    @override
    def __len__(self) -> int:
        registry = self._get_registry()
        return registry.__len__()

    @override
    def add(self, key: K, instance: T) -> bool:
        registry = self._get_registry()
        return registry.add(key, instance)

    @override
    def has(self, key: K) -> bool:
        registry = self._get_registry()
        return registry.has(key)

    @override
    def get(self, key: K) -> Maybe[T]:
        registry = self._get_registry()
        return registry.get(key)

    @override
    def items(self, filter_fn: Callable[[K], bool] | None = None) -> Iterable[tuple[K, T]]:
        registry = self._get_registry()
        yield from registry.items(filter_fn)

    @override
    def clear(self) -> None:
        registry = self._get_registry()
        registry.clear()
        del self._storage[self._scope_func()]


class ScopedMutDictRegistry[K, T](MutDictRegistry[K, T]):
    def __init__(
        self, factory: Callable[[], MutDictRegistry[K, T]], scope_func: Callable[[], str]
    ) -> None:
        super().__init__()
        self._scope_func = scope_func
        self._storage: dict[str, MutDictRegistry[K, T]] = defaultdict(factory)

    def _get_registry(self) -> MutDictRegistry[K, T]:
        key = self._scope_func()
        return self._storage[key]

    @override
    def __len__(self) -> int:
        registry = self._get_registry()
        return registry.__len__()

    @override
    def add(self, key: K, instance: T) -> bool:
        registry = self._get_registry()
        return registry.add(key, instance)

    @override
    def has(self, key: K) -> bool:
        registry = self._get_registry()
        return registry.has(key)

    @override
    def get(self, key: K) -> Maybe[T]:
        registry = self._get_registry()
        return registry.get(key)

    @override
    def items(self, filter_fn: Callable[[K], bool] | None = None) -> Iterable[tuple[K, T]]:
        registry = self._get_registry()
        yield from registry.items(filter_fn)

    @override
    def remove(self, key: K, instance: T | None = None) -> bool:
        registry = self._get_registry()
        return registry.remove(key, instance)

    @override
    def clear(self) -> None:
        registry = self._get_registry()
        registry.clear()
        del self._storage[self._scope_func()]


class NamedDictRegistry[K, T](NamedCollector[K, T], NamedListSelector[K, T]):
    __slots__ = ('_storage',)

    def __init__(self) -> None:
        self._storage: dict[K, dict[str | None, T]] = defaultdict(dict)

    def __len__(self) -> int:
        return len(self._storage)

    @override
    def add(self, key: K, instance: T, name: str | None = None) -> bool:
        if key in self._storage and name in self._storage[key]:
            return instance is self._storage[key][name]
        self._storage[key][name] = instance
        return True

    @override
    def has(self, key: K, name: str | None = None) -> bool:
        if key not in self._storage:
            return False
        if name is None:
            return True
        return name in self._storage[key]

    @override
    def get(self, key: K, name: str | None = None) -> Maybe[T]:
        if key not in self._storage:
            return Nothing

        if name is None:
            first_name = next(iter(self._storage[key]))
            return Just(self._storage[key][first_name])
        return Just(self._storage[key][name])

    @override
    def items(
        self, filter_fn: Callable[[K, str], bool] | None = None
    ) -> Iterable[tuple[K, str, T]]:
        for k, values in self._storage.items():
            for name, v in values.items():
                if not filter_fn or filter_fn(k, name):
                    yield k, name, v

    def clear(self) -> None:
        while key := next(iter(self._storage), None):
            del self._storage[key]


class MutNamedDictRegistry[K, T](MutNamedCollector[K, T], NamedDictRegistry[K, T]):
    @override
    def remove(self, key: K, instance: T | None = None, name: str | None = None) -> bool:
        if key not in self._storage:
            return False

        if name is None:
            values = self._storage[key]

            if instance is None:
                del self._storage[key]
            else:
                while len(values):
                    name = next(iter(values))
                    if instance == values[name]:
                        del values[name]
                if len(values) == 0:
                    del self._storage[key]

            return True

        if name not in self._storage[key]:
            return False

        del self._storage[key][name]

        if len(self._storage[key]) == 0:
            del self._storage[key]
        return True



from collections import defaultdict
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, override

from greyhorse.logging import logger
from greyhorse.maybe import Maybe, Nothing
from greyhorse.result import Err, Ok, Result
from greyhorse.utils.injectors import ParamsInjector
from greyhorse.utils.invoke import invoke_sync

from ..abc.component import Component, ComponentError
from ..abc.controllers import Controller, ControllerError, ControllerFactoryFn
from ..abc.operators import Operator
from ..abc.providers import Provider
from ..abc.resources import Lifetime
from ..abc.services import Service, ServiceError, ServiceFactoryFn
from ..abc.visitor import Visitor
from ..registries import MutDictRegistry
from ..resources import (
    Container,
    ResourceManager,
    inject_targets,
    make_container,
    uninject_targets,
)
from ..schemas.components import ComponentConf, ModuleComponentConf
from ..schemas.elements import CtrlConf, SvcConf

if TYPE_CHECKING:
    from .module import Module


class SyncComponent(Component):
    def __init__(self, name: str, path: str, conf: ComponentConf) -> None:
        super().__init__(name, path)
        self._conf = conf
        self._rm = ResourceManager()

        self._controllers: list[Controller] = []
        self._services: list[Service] = []

        self._resources = MutDictRegistry[type, Any]()
        self._container: Container | None = None
        self._providers = MutDictRegistry[type[Provider], Provider]()
        self._operators: dict[type, list[Operator]] = defaultdict(list)

    @override
    def accept_visitor(self, visitor: Visitor) -> None:
        visitor.start_component(self)
        for svc in self._services:
            svc.accept_visitor(visitor)
        for ctrl in self._controllers:
            ctrl.accept_visitor(visitor)
        visitor.finish_component(self)

    @override
    def get_provider[P: Provider](self, prov_type: type[P]) -> Maybe[P]:
        if prov_type in self._conf.providers:
            return self._container.get(prov_type)
        return Nothing

    @override
    def get_operators[T](self, res_type: type[T]) -> Iterable[Operator[T]]:
        if res_type in self._conf.operators:
            return self._operators[res_type].copy()
        return []

    @override
    def add_resource[T](self, res_type: type[T], resource: T) -> bool:
        if res_type in self._conf.resource_claims:
            return self._resources.add(res_type, resource)
        return False

    @override
    def remove_resource[T](self, res_type: type[T]) -> bool:
        if res_type in self._conf.resource_claims:
            return self._resources.remove(res_type)
        return False

    def add_controller(self, controller: Controller) -> bool:
        if self._controllers.count(controller):
            return False
        self._controllers.append(controller)
        return self._rm.add_controller(controller)

    def remove_controller(self, controller: Controller) -> bool:
        if not self._controllers.count(controller):
            return False
        self._controllers.remove(controller)
        return self._rm.remove_controller(controller)

    def add_service(self, service: Service) -> bool:
        if self._services.count(service):
            return False
        self._services.append(service)
        return self._rm.add_service(service)

    def remove_service(self, service: Service) -> bool:
        if not self._services.count(service):
            return False
        self._services.remove(service)
        return self._rm.remove_service(service)

    @override
    def create(self) -> Result[None, ComponentError]:
        logger.info('{path}: Component "{name}" create'.format(path=self._path, name=self.name))

        self._container = make_container(lifetime=Lifetime.COMPONENT())

        inject_targets(
            self._container,
            [svc.type_.__module__ for svc in self._conf.services]
            + [ctrl.type_.__module__ for ctrl in self._conf.controllers],
        )

        injector = ParamsInjector()

        if not (res := self._create_services(injector)):
            return res  # type: ignore

        for svc in res.unwrap():
            injector.add_type_provider(type(svc), svc)
            self.add_service(svc)

        if not (res := self._create_controllers(injector)):
            return res  # type: ignore

        for svc in self._services:
            injector.remove_type_provider(type(svc))

        for ctrl in res.unwrap():
            self.add_controller(ctrl)

        self._rm.install_container(self._container)

        if not (
            res := self._rm.setup(self._providers).map_err(
                lambda e: ComponentError.Resource(
                    path=self._path, name=self.name, details=e.message
                )
            )
        ):
            return res

        for op in self._rm.get_operators():
            self._operators[op.wrapped_type].append(op)

        logger.info(
            '{path}: Component "{name}" create successful'.format(
                path=self._path, name=self.name
            )
        )

        return Ok()

    @override
    def setup(self) -> Result[None, ComponentError]:
        logger.info('{path}: Component "{name}" setup'.format(path=self._path, name=self.name))

        for k, v in self._resources.items():
            self._container.add_resource(k, v)

        self._container.context.__enter__()

        injector = ParamsInjector()
        injector.add_type_provider(Container, self._container)

        for ctrl, _ctrl_conf in zip(self._controllers, self._conf.controllers, strict=False):
            injected_args = injector(ctrl.setup)

            if not (
                res := invoke_sync(
                    ctrl.setup, *injected_args.args, **injected_args.kwargs
                ).map_err(
                    lambda e: ComponentError.Ctrl(
                        path=self._path, name=self.name, details=e.message
                    )
                )
            ):
                return res

        injector.remove_type_provider(Container)

        for svc, _svc_conf in zip(self._services, self._conf.services, strict=False):
            injected_args = injector(svc.setup)

            if not (
                res := invoke_sync(
                    svc.setup, *injected_args.args, **injected_args.kwargs
                ).map_err(
                    lambda e: ComponentError.Service(
                        path=self._path, name=self.name, details=e.message
                    )
                )
            ):
                return res

        logger.info(
            '{path}: Component "{name}" setup successful'.format(
                path=self._path, name=self.name
            )
        )

        return Ok()

    @override
    def teardown(self) -> Result[None, ComponentError]:
        logger.info(
            '{path}: Component "{name}" teardown'.format(path=self._path, name=self.name)
        )

        injector = ParamsInjector()

        for svc, _svc_conf in zip(
            reversed(self._services), reversed(self._conf.services), strict=False
        ):
            injected_args = injector(svc.teardown)

            if not (
                res := invoke_sync(
                    svc.teardown, *injected_args.args, **injected_args.kwargs
                ).map_err(
                    lambda e: ComponentError.Service(
                        path=self._path, name=self.name, details=e.message
                    )
                )
            ):
                return res

        injector.add_type_provider(Container, self._container)

        for ctrl, _ctrl_conf in zip(
            reversed(self._controllers), reversed(self._conf.controllers), strict=False
        ):
            injected_args = injector(ctrl.teardown)

            if not (
                res := invoke_sync(
                    ctrl.teardown, *injected_args.args, **injected_args.kwargs
                ).map_err(
                    lambda e: ComponentError.Ctrl(
                        path=self._path, name=self.name, details=e.message
                    )
                )
            ):
                return res

        injector.remove_type_provider(Container)

        self._container.context.__exit__()

        for k, _v in self._resources.items():  # noqa: PERF102
            self._container.remove_resource(k)

        logger.info(
            '{path}: Component "{name}" teardown successful'.format(
                path=self._path, name=self.name
            )
        )

        return Ok()

    @override
    def destroy(self) -> Result[None, ComponentError]:
        logger.info(
            '{path}: Component "{name}" destroy'.format(path=self._path, name=self.name)
        )

        for op in self._rm.get_operators():
            self._operators[op.wrapped_type].remove(op)

        if not (
            res := self._rm.teardown().map_err(
                lambda e: ComponentError.Resource(
                    path=self._path, name=self.name, details=e.message
                )
            )
        ):
            return res

        while ctrl := next(reversed(self._controllers), None):
            self.remove_controller(ctrl)

        while svc := next(reversed(self._services), None):
            self.remove_service(svc)

        assert not self._controllers
        assert not self._services

        uninject_targets(
            [svc.type_.__module__ for svc in self._conf.services]
            + [ctrl.type_.__module__ for ctrl in self._conf.controllers]
        )

        self._container = None

        logger.info(
            '{path}: Component "{name}" destroy successful'.format(
                path=self._path, name=self.name
            )
        )

        return Ok()

    def _create_controller(
        self, conf: CtrlConf, factory: ControllerFactoryFn, injector: ParamsInjector
    ) -> Result[Controller, ControllerError]:
        logger.info(
            '{path}: Component controller "{name}" create'.format(
                path=self._path, name=conf.name
            )
        )

        values = conf.args.copy()
        values['name'] = conf.name
        injected_args = injector(factory, values=values)

        try:
            if not (res := factory(*injected_args.args, **injected_args.kwargs)):
                return res

        except Exception as e:
            error = ControllerError.Factory(details=str(e))
            logger.error(error.message)
            return error.to_result()

        logger.info(
            '{path}: Component controller "{name}" created successfully'.format(
                path=self._path, name=conf.name
            )
        )

        if isinstance(res, Controller):
            return Ok(res)

        return res

    def _create_service(
        self, conf: SvcConf, factory: ServiceFactoryFn, injector: ParamsInjector
    ) -> Result[Service, ServiceError]:
        logger.info(
            '{path}: Component service "{name}" create'.format(path=self._path, name=conf.name)
        )

        values = conf.args.copy()
        values['name'] = conf.name
        injected_args = injector(factory, values=values)

        try:
            if not (res := factory(*injected_args.args, **injected_args.kwargs)):
                return res

        except Exception as e:
            error = ServiceError.Factory(details=str(e))
            logger.error(error.message)
            return error.to_result()

        logger.info(
            '{path}: Component service "{name}" created successfully'.format(
                path=self._path, name=conf.name
            )
        )

        if isinstance(res, Service):
            return Ok(res)

        return res

    def _create_controllers(
        self, injector: ParamsInjector
    ) -> Result[list[Controller], ComponentError]:
        result = []

        for conf in self._conf.controllers:
            factory = self._conf.controller_factories.get(conf.type_, conf.type_)

            match self._create_controller(conf, factory, injector):
                case Ok(ctrl):
                    result.append(ctrl)

                case Err(e):
                    error = ComponentError.Ctrl(
                        path=self._path, name=self.name, details=e.message
                    )
                    logger.error(error.message)
                    return error.to_result()

        return Ok(result)

    def _create_services(
        self, injector: ParamsInjector
    ) -> Result[list[Service], ComponentError]:
        result = []

        for conf in self._conf.services:
            factory = self._conf.service_factories.get(conf.type_, conf.type_)

            match self._create_service(conf, factory, injector):
                case Ok(svc):
                    result.append(svc)

                case Err(e):
                    error = ComponentError.Service(
                        path=self._path, name=self.name, details=e.message
                    )
                    logger.error(error.message)
                    return error.to_result()

        return Ok(result)


class SyncModuleComponent(SyncComponent):
    def __init__(
        self, name: str, path: str, conf: ModuleComponentConf, module: 'Module'
    ) -> None:
        super().__init__(name, path, conf)
        self._conf = conf
        self._module = module

    @override
    def accept_visitor(self, visitor: Visitor) -> None:
        visitor.start_component(self)
        for svc in self._services:
            svc.accept_visitor(visitor)
        for ctrl in self._controllers:
            ctrl.accept_visitor(visitor)
        self._module.accept_visitor(visitor)
        visitor.finish_component(self)

    @override
    def create(self) -> Result[None, ComponentError]:
        if not (res := super().create()):
            return res

        for prov_type in self._module.conf.provider_claims:
            if prov := self._rm.find_provider(prov_type, self._providers).unwrap_or_none():
                self._module.add_provider(prov_type, prov)

        if not (
            res := self._module.create().map_err(
                lambda e: ComponentError.Module(
                    path=self._path, name=self.name, details=e.message
                )
            )
        ):
            return res

        return Ok()

    @override
    def setup(self) -> Result[None, ComponentError]:
        if not (res := super().setup()):
            return res

        for res_type in self._module.conf.resource_claims:
            if res := self._container.get(res_type).unwrap_or_none():
                self._module.add_resource(res_type, res)

        for res_type in self._module.conf.operators:
            for op in self._operators[res_type]:
                self._module.add_operator(op)

        if not (
            res := self._module.setup().map_err(
                lambda e: ComponentError.Module(
                    path=self._path, name=self.name, details=e.message
                )
            )
        ):
            return res

        for prov_type in self._module.conf.providers:
            if prov := self._module.get_provider(prov_type).unwrap_or_none():
                self._providers.add(prov_type, prov)

        return Ok()

    @override
    def teardown(self) -> Result[None, ComponentError]:
        for prov_type in reversed(self._module.conf.providers):
            self._providers.remove(prov_type)

        if not (
            res := self._module.teardown().map_err(
                lambda e: ComponentError.Module(
                    path=self._path, name=self.name, details=e.message
                )
            )
        ):
            return res

        for res_type in reversed(self._module.conf.operators):
            for op in self._operators[res_type]:
                self._module.remove_operator(op)

        for res_type in reversed(self._module.conf.resource_claims):
            self._module.remove_resource(res_type)

        return super().teardown()

    @override
    def destroy(self) -> Result[None, ComponentError]:
        if not (
            res := self._module.destroy().map_err(
                lambda e: ComponentError.Module(
                    path=self._path, name=self.name, details=e.message
                )
            )
        ):
            return res

        for prov_type in reversed(self._module.conf.provider_claims):
            self._module.remove_provider(prov_type)

        return super().destroy()



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .component import Component
    from .controllers import Controller
    from .module import Module
    from .services import Service


class Visitor:
    def visit_controller(self, controller: 'Controller') -> None: ...

    def visit_service(self, service: 'Service') -> None: ...

    def start_component(self, component: 'Component') -> None: ...

    def finish_component(self, component: 'Component') -> None: ...

    def start_module(self, module: 'Module') -> None: ...

    def finish_module(self, module: 'Module') -> None: ...



import asyncio
import threading
from asyncio import Future as AsyncFuture
from asyncio import iscoroutine
from collections.abc import Callable
from dataclasses import dataclass
from queue import Empty, Queue
from types import TracebackType
from typing import TYPE_CHECKING, Any

from greyhorse.utils.types import is_awaitable

if TYPE_CHECKING:
    from greyhorse.app.resources import Container


@dataclass(slots=True, frozen=True)
class _Task:
    func: Callable[[...], Any]
    args: tuple
    kwargs: dict[str, Any]
    waiter: AsyncFuture


class Runtime:
    __slots__ = ('_loop', '_thread', '_counter', '_tasks', '_sync_ctx', '_container')

    _instance: 'Runtime' = None

    def __new__(cls) -> 'Runtime':
        if not isinstance(cls._instance, cls):
            from greyhorse.app.abc.resources import Lifetime
            from greyhorse.app.resources import make_container

            container = make_container(lifetime=Lifetime.RUNTIME())
            cls._instance = object.__new__(cls)
            cls._instance._loop = asyncio.new_event_loop()  # noqa: SLF001
            cls._instance._thread = threading.Thread(  # noqa: SLF001
                name='Greyhorse runtime',
                target=cls._instance._loop.run_forever,  # noqa: SLF001
            )
            cls._instance._counter = 0  # noqa: SLF001
            cls._instance._tasks = Queue()  # noqa: SLF001
            cls._instance._sync_ctx = False  # noqa: SLF001
            cls._instance._container = container  # noqa: SLF001

        return cls._instance

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    @property
    def active(self) -> bool:
        return self._thread.is_alive()

    @property
    def container(self) -> 'Container':
        return self._container

    def start(self) -> None:
        if self._counter == 0:
            self._thread.start()
        self._counter += 1

    def stop(self) -> None:
        if self._counter == 1:
            self._loop.call_soon_threadsafe(self._loop.stop)
            self._thread.join()
            self._loop.close()

        self._counter = max(self._counter - 1, 0)

    def __enter__(self) -> 'Container':
        self.start()
        return self._container.context.__enter__()

    def __exit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        try:
            self._container.context.__exit__(exc_type, exc_value, traceback)
        finally:
            self.stop()

    def invoke_sync[T, **P](
        self, func: Callable[P, T], /, *args: P.args, **kwargs: P.kwargs
    ) -> T:
        if is_awaitable(func):
            if self.active:
                future = asyncio.run_coroutine_threadsafe(func(*args, **kwargs), self._loop)
                self._wait_for_async_future(future)
                return future.result()
            return asyncio.run(func(*args, **kwargs))

        return func(*args, **kwargs)

    async def invoke_async[T, **P](
        self, func: Callable[P, T], /, *args: P.args, **kwargs: P.kwargs
    ) -> T:
        if is_awaitable(func):
            if iscoroutine(func):
                return await func  # type: ignore
            return await func(*args, **kwargs)

        if self._sync_ctx:
            future = self._loop.create_future()
            task = _Task(func=func, args=args, kwargs=kwargs, waiter=future)
            self._tasks.put(task)
            return await future

        return await asyncio.to_thread(func, *args, **kwargs)

    def _wait_for_async_future(self, future: AsyncFuture) -> None:
        self._sync_ctx = not future.done()
        while self._sync_ctx:
            try:
                while task := self._tasks.get(timeout=0.00001):
                    try:
                        res = task.func(*task.args, **task.kwargs)
                        self._loop.call_soon_threadsafe(task.waiter.set_result, res)

                    except Exception as e:
                        self._loop.call_soon_threadsafe(task.waiter.set_exception, e)

                    self._tasks.task_done()

            except Empty:
                self._sync_ctx = not future.done()



from __future__ import annotations

from greyhorse.app.abc.providers import FactoryProvider
from greyhorse.app.builders.module import ModuleBuilder
from greyhorse.app.schemas.components import ModuleComponentConf, ModuleConf
from greyhorse.app.schemas.elements import CtrlConf, SvcConf

from .root import FunctionalOperatorCtrl, FunctionalOperatorService
from .testmodule.common.functional import FunctionalOperator


def __init__() -> ModuleConf:  # noqa: N807
    return ModuleConf(
        enabled=True,
        operators=[FunctionalOperator],
        providers=[FactoryProvider[FunctionalOperator]],
        components={
            'dict': ModuleComponentConf(
                enabled=True,
                path='..testmodule.module',
                providers=[FactoryProvider[FunctionalOperator]],
                services=[SvcConf(type=FunctionalOperatorService)],
                controllers=[CtrlConf(type=FunctionalOperatorCtrl)],
            )
        },
    )


def test_module() -> None:
    module_conf = __init__()

    module_builder = ModuleBuilder(module_conf, 'tests')
    res = module_builder.create_pass()
    assert res.is_ok()
    module = res.unwrap()

    res = module.create()
    assert res.is_ok()

    res = module.setup()
    assert res.is_ok()

    prov = module.get_provider(FactoryProvider[FunctionalOperator]).unwrap()

    print('OK')

    op = prov.create().unwrap()

    res = op.add_number(123)
    assert res.is_ok()

    res = op.get_number()
    assert res.is_ok()
    assert res.unwrap() == 123

    res = op.remove_number()
    assert res.is_ok()
    assert res.unwrap() is True

    res = op.get_number()
    assert res.is_err()
    assert res.unwrap_err() == 'Number is not initialized'

    prov.destroy(op)

    res = module.teardown()
    assert res.is_ok()

    res = module.destroy()
    assert res.is_ok()



from collections import OrderedDict, defaultdict
from collections.abc import Callable, Iterable
from functools import partial
from types import AsyncGeneratorType, GeneratorType

import networkx as nx

from greyhorse.error import Error, ErrorCase
from greyhorse.result import Err, Ok, Result
from greyhorse.utils.invoke import invoke_sync

from ..abc.controllers import Controller, OperatorMember
from ..abc.operators import Operator
from ..abc.providers import (
    FactoryProvider,
    ForwardProvider,
    MutProvider,
    Provider,
    SharedProvider,
)
from ..abc.selectors import Selector
from ..abc.services import ResourceMember, Service
from ..boxes import (
    AsyncFactoryGenBox,
    AsyncForwardGenBox,
    AsyncMutGenBox,
    AsyncSharedGenBox,
    SyncFactoryGenBox,
    SyncForwardGenBox,
    SyncMutGenBox,
    SyncSharedGenBox,
)
from . import Container
from .injection import _invoke_target
from .mappers import SyncResourceMapper


class ResourceError(Error):
    namespace = 'greyhorse.app'

    Provision = ErrorCase(msg='Resource provision error: "{details}"', details=str)
    NoSuchResource = ErrorCase(msg='No such resource: "{resource}"', resource=str)
    NoSuchDependency = ErrorCase(msg='No such dependency: "{resource}"', resource=str)


class ResourceManager:
    def __init__(self) -> None:
        self._services: list[Service] = []
        self._controllers: list[Controller] = []
        self._deps = set()

        self._resource_graph = nx.DiGraph()
        self._operator_map: dict[type, dict[Controller, OperatorMember]] = defaultdict(dict)
        self._provided_resources = OrderedDict[Operator, SyncResourceMapper]()
        self._public_operators: list[Operator] = []

    def get_operators(self) -> Iterable[Operator]:
        return self._public_operators.copy()

    def add_service(self, service: Service) -> bool:
        if self._services.count(service):
            return False
        self._services.append(service)
        service.inspect(partial(self._add_service_resource, service))
        return True

    def remove_service(self, service: Service) -> bool:
        if not self._services.count(service):
            return False
        service.inspect(partial(self._remove_service_resource, service))
        self._services.remove(service)
        return True

    def add_controller(self, controller: Controller) -> bool:
        if self._controllers.count(controller):
            return False
        self._controllers.append(controller)
        controller.inspect(partial(self._add_ctrl_operator, controller))
        return True

    def remove_controller(self, controller: Controller) -> bool:
        if not self._controllers.count(controller):
            return False
        controller.inspect(partial(self._remove_ctrl_operator, controller))
        self._controllers.remove(controller)
        return True

    def setup(
        self, providers: Selector[type[Provider], Provider] | None = None
    ) -> Result[None, ResourceError]:
        operators = []

        for ctrl_dict in self._operator_map.values():
            for ctrl, operator_member in ctrl_dict.items():
                factory = partial(operator_member.method, self=ctrl)
                match invoke_sync(factory):
                    case Ok(op) | (Operator() as op):
                        operators.append(op)

                    case Err(e):
                        return ResourceError.Provision(details=e).to_result()

                    case _:
                        continue

        for op in operators:
            if not self.setup_operator(op, providers=providers):
                self._public_operators.append(op)

        return Ok()

    def teardown(self) -> Result[None, ResourceError]:
        while operator := next(reversed(self._provided_resources.keys()), None):
            mapper = self._provided_resources[operator]

            if not (
                res := mapper.teardown().map_err(
                    lambda e: ResourceError.Provision(details=e.message)
                )
            ):
                return res

            del self._provided_resources[operator]

        self._public_operators = []
        self._deps.clear()
        self._resource_graph.clear()
        assert len(self._provided_resources) == 0
        return Ok()

    @staticmethod
    def _invoke_member(
        container: Container, member: ResourceMember, fn: Callable[[], object], _: object
    ) -> object:
        factory = partial(_invoke_target, fn, member.params, container) if member.params else fn

        if issubclass(member.resource_type, Provider):
            prov = None

            match invoke_sync(factory):
                case Ok(prov) | (Provider() as prov):
                    pass

                case GeneratorType() as gen:
                    if issubclass(member.resource_type, SharedProvider):
                        prov = SyncSharedGenBox[member.resource_type.__wrapped_type__](factory)
                    elif issubclass(member.resource_type, MutProvider):
                        prov = SyncMutGenBox[member.resource_type.__wrapped_type__](factory)
                    elif issubclass(member.resource_type, FactoryProvider):
                        prov = SyncFactoryGenBox[member.resource_type.__wrapped_type__](factory)
                    elif issubclass(member.resource_type, ForwardProvider):
                        prov = SyncForwardGenBox[member.resource_type.__wrapped_type__](gen)

                case AsyncGeneratorType() as gen:
                    if issubclass(member.resource_type, SharedProvider):
                        prov = AsyncSharedGenBox[member.resource_type.__wrapped_type__](factory)
                    elif issubclass(member.resource_type, MutProvider):
                        prov = AsyncMutGenBox[member.resource_type.__wrapped_type__](factory)
                    elif issubclass(member.resource_type, FactoryProvider):
                        prov = AsyncFactoryGenBox[member.resource_type.__wrapped_type__](
                            factory
                        )
                    elif issubclass(member.resource_type, ForwardProvider):
                        prov = AsyncForwardGenBox[member.resource_type.__wrapped_type__](gen)

                case _:
                    raise AssertionError(
                        'Unexpected return value returned from provider method'
                    )

            return prov

        return factory()

    def install_container(self, container: Container) -> bool:
        result = True

        for resource_type, node_data in self._resource_graph.nodes(data=True):
            if 'member' not in node_data:
                continue

            member: ResourceMember = node_data['member']
            fn = partial(self._invoke_member, container, member, node_data['factory'])
            registry = container.child_registry(member.lifetime).unwrap_or(container.registry)
            result &= registry.add_factory(resource_type, fn, cache=member.cache)

        return result

    def setup_operator[T](
        self,
        operator: Operator[T],
        container: Container | None = None,
        providers: Selector[type[Provider], Provider] | None = None,
    ) -> Result[bool, ResourceError]:
        res_type = operator.wrapped_type
        prov_type = None

        if operator in self._provided_resources:
            return Ok(False)

        if issubclass(res_type, Provider):
            prov_type = res_type
        else:
            found = False

            candidates = (
                SharedProvider[res_type],
                ForwardProvider[res_type],
                FactoryProvider[res_type],
                MutProvider[res_type],
            )

            for prov_type in candidates:
                if self._resource_graph.has_node(prov_type) or (
                    providers.has(prov_type) if providers is not None else False
                ):
                    found = True
                    break

            if not found:
                return ResourceError.NoSuchResource(resource=res_type.__name__).to_result()

        prov = container.get(prov_type).unwrap_or_none() if container is not None else None
        if prov is None:
            prov = providers.get(prov_type).unwrap_or_none() if providers is not None else None
        if prov is None:
            return ResourceError.NoSuchResource(resource=res_type.__name__).to_result()

        mapper = SyncResourceMapper[prov_type](prov, operator)

        if res := mapper.setup():
            self._provided_resources[operator] = mapper

        return res.map(lambda _: True).map_err(lambda e: ResourceError.Provision(details=e))

    def teardown_operator[T](self, operator: Operator[T]) -> Result[bool, ResourceError]:
        if not (mapper := self._provided_resources.get(operator)):
            return Ok(False)

        if res := mapper.teardown():
            del self._provided_resources[operator]

        return res.map(lambda _: True).map_err(lambda e: ResourceError.Provision(details=e))

    def _add_service_resource(self, service: Service, member: ResourceMember) -> None:
        if self._resource_graph.has_node(member.resource_type):
            return

        self._resource_graph.add_node(
            member.resource_type, factory=partial(member.method, self=service), member=member
        )

        for param_type in member.params.values():
            if not self._resource_graph.has_node(param_type):
                self._deps.add(param_type)

            self._resource_graph.add_edge(member.resource_type, param_type)

    def _remove_service_resource(self, _: Service, member: ResourceMember) -> None:
        if not self._resource_graph.has_node(member.resource_type):
            return

        self._resource_graph.remove_node(member.resource_type)

        for param_type in member.params.values():
            if not self._resource_graph.has_node(param_type):
                self._deps.remove(param_type)

    def _add_ctrl_operator(
        self, controller: Controller, operator_member: OperatorMember
    ) -> None:
        ctrl_dict = self._operator_map[operator_member.resource_type]

        if controller in ctrl_dict:
            return

        ctrl_dict[controller] = operator_member

    def _remove_ctrl_operator(
        self, controller: Controller, operator_member: OperatorMember
    ) -> None:
        ctrl_dict = self._operator_map[operator_member.resource_type]

        if controller not in ctrl_dict:
            return

        del ctrl_dict[controller]



import inspect
from asyncio import AbstractEventLoop, get_running_loop
from collections.abc import Callable

from ..app.runtime import Runtime


def is_like_sync_context_manager(instance: object) -> bool:
    if callable(instance) and inspect.isgeneratorfunction(inspect.unwrap(instance)):
        instance = instance()
    return hasattr(instance, '__enter__') and hasattr(instance, '__exit__')


def is_like_async_context_manager(instance: object) -> bool:
    if callable(instance) and inspect.isasyncgenfunction(inspect.unwrap(instance)):
        instance = instance()
    return hasattr(instance, '__aenter__') and hasattr(instance, '__aexit__')


def get_asyncio_loop() -> AbstractEventLoop | None:
    try:
        return get_running_loop()
    except RuntimeError:
        return None


def invoke_sync[T, **P](func: Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
    return Runtime().invoke_sync(func, *args, **kwargs)


async def invoke_async[T, **P](func: Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
    return await Runtime().invoke_async(func, *args, **kwargs)


def caller_path(depth: int) -> list[str]:
    import inspect

    frame = inspect.currentframe()

    for _ in range(depth, 0, -1):
        frame = frame.f_back

    return inspect.getmodule(frame).__name__.split('.')



from abc import ABC, abstractmethod
from collections.abc import Iterable

from greyhorse.error import Error, ErrorCase
from greyhorse.maybe import Maybe
from greyhorse.result import Result

from .operators import Operator
from .providers import Provider
from .visitor import Visitor


class ComponentError(Error):
    namespace = 'greyhorse.app.component'

    Resource = ErrorCase(
        msg='{path}: Resource error in component: "{name}", details: "{details}"',
        path=str,
        name=str,
        details=str,
    )
    Ctrl = ErrorCase(
        msg='{path}: Controller error in component: "{name}", details: "{details}"',
        path=str,
        name=str,
        details=str,
    )
    Service = ErrorCase(
        msg='{path}: Service error in component: "{name}", details: "{details}"',
        path=str,
        name=str,
        details=str,
    )
    Module = ErrorCase(
        msg='{path}: Submodule error in component: "{name}", details: "{details}"',
        path=str,
        name=str,
        details=str,
    )


class Component(ABC):
    __slots__ = ('_name', '_path')

    def __init__(self, name: str, path: str) -> None:
        self._name = name
        self._path = path

    @property
    def name(self) -> str:
        return self._name

    @property
    def path(self) -> str:
        return self._path

    @abstractmethod
    def accept_visitor(self, visitor: Visitor) -> None: ...

    @abstractmethod
    def get_provider[P: Provider](self, prov_type: type[P]) -> Maybe[P]: ...

    @abstractmethod
    def get_operators[T](self, res_type: type[T]) -> Iterable[Operator[T]]: ...

    @abstractmethod
    def add_resource[T](self, res_type: type[T], resource: T) -> bool: ...

    @abstractmethod
    def remove_resource[T](self, res_type: type[T]) -> bool: ...

    @abstractmethod
    def create(self) -> Result[None, ComponentError]: ...

    @abstractmethod
    def setup(self) -> Result[None, ComponentError]: ...

    @abstractmethod
    def teardown(self) -> Result[None, ComponentError]: ...

    @abstractmethod
    def destroy(self) -> Result[None, ComponentError]: ...



from .container import Container, make_container, root  # noqa: F401
from .factories import TypeFactory  # noqa: F401
from .injection import inject, inject_targets, uninject_targets  # noqa: F401
from .manager import ResourceError, ResourceManager  # noqa: F401
from .mappers import AsyncResourceMapper, ResourceMapper, SyncResourceMapper  # noqa: F401
from .registry import FactoryRegistry  # noqa: F401



from greyhorse.app.abc.providers import FactoryProvider
from greyhorse.app.entities.application import Application
from greyhorse.app.schemas.components import ModuleComponentConf, ModuleConf
from greyhorse.app.schemas.elements import CtrlConf, SvcConf

from .root import FunctionalOperatorCtrl, FunctionalOperatorService
from .testmodule.common.functional import FunctionalOperator


def test_app() -> None:
    app_conf = ModuleConf(
        enabled=True,
        providers=[FactoryProvider[FunctionalOperator]],
        components={
            'dict': ModuleComponentConf(
                enabled=True,
                path='..testmodule.module',
                providers=[FactoryProvider[FunctionalOperator]],
                services=[SvcConf(type=FunctionalOperatorService)],
                controllers=[CtrlConf(type=FunctionalOperatorCtrl)],
            )
        },
    )

    app = Application('TestApp')

    res = app.load(app_conf)
    assert res.is_ok()

    res = app.setup()
    assert res.is_ok()

    print('OK')

    prov = app.get_provider(FactoryProvider[FunctionalOperator])
    assert prov.is_just()

    prov = prov.unwrap()
    op = prov.create().unwrap()

    res = op.add_number(123)
    assert res.is_ok()

    res = op.get_number()
    assert res.is_ok()
    assert res.unwrap() == 123

    res = op.remove_number()
    assert res.is_ok()
    assert res.unwrap() is True

    res = op.get_number()
    assert res.is_err()
    assert res.unwrap_err() == 'Number is not initialized'

    assert app.start()

    # with app.graceful_exit():
    # app.run_sync()
    # await app.run_async()

    assert app.stop()

    prov.destroy(op)

    res = app.teardown()
    assert res.is_ok()

    res = app.unload()
    assert res.is_ok()



import inspect
import sys
from collections.abc import Callable, Iterable
from dataclasses import dataclass, field
from functools import wraps
from typing import Any, TypeVar, get_type_hints

from greyhorse.maybe import Just, Maybe, Nothing
from greyhorse.utils.invoke import caller_path
from greyhorse.utils.types import is_maybe, is_optional, unwrap_maybe, unwrap_optional

from ..abc.module import Module
from .container import Container, root

Target = TypeVar('Target', bound=Callable | type)


@dataclass(slots=True)
class _ModuleNode:
    children: dict[str, '_ModuleNode'] = field(default_factory=dict)
    targets: list[Target] = field(default_factory=list)


class _TargetCollector:
    __slots__ = ('_root',)

    def __init__(self) -> None:
        self._root = _ModuleNode()

    def add_target(self, target: Target) -> bool:
        path = target.__module__
        current = self._root

        for path_entry in path.split('.'):
            if path_entry not in current.children:
                current.children[path_entry] = _ModuleNode()
            current = current.children[path_entry]

        if target in current.targets:
            return False

        current.targets.append(target)
        return True

    def list_targets(self, path: str) -> list[Target]:
        current = self._root

        for path_entry in path.split('.'):
            if path_entry not in current.children:
                return []
            current = current.children[path_entry]

        result = current.targets.copy()
        nodes = list(current.children.values())
        curr_idx = 0

        while curr_idx < len(nodes):
            node = nodes[curr_idx]
            nodes += node.children.values()
            curr_idx += 1

        for node in nodes:
            result += node.targets.copy()

        return result


_collector = _TargetCollector()


def get_current_module(depth: int = 2) -> Maybe[Module]:
    current_module_path = caller_path(depth)

    while len(current_module_path):
        dotted_module_path = '.'.join(current_module_path)
        if (mod := sys.modules.get(dotted_module_path)) and (
            instance := getattr(mod, '__gh_module__', None)
        ):
            return Just(instance)

        current_module_path = current_module_path[: len(current_module_path) - 1]

    return Nothing


def inject_targets(container: Container, paths: Iterable[str]) -> int:
    targets = set()

    for path in paths:
        targets.update(set(_collector.list_targets(path)))

    for target in targets:
        target.__container__ = container

    return len(targets)


def uninject_targets(paths: Iterable[str]) -> int:
    targets = set()

    for path in paths:
        targets.update(set(_collector.list_targets(path)))

    for target in targets:
        delattr(target, '__container__')

    return len(targets)


def _invoke_target[T, **P](
    func: Callable[P, T],
    hints: dict[str, Any],
    container: Container,
    /,
    *args: P.args,
    **kwargs: P.kwargs,
) -> T:
    injected_args = {}
    args_count = len(args) if inspect.ismethod(func) else len(args) - 1

    for i, (k, v) in enumerate(hints.items()):
        if i < args_count or k in kwargs:
            continue
        if value := container.get(unwrap_maybe(unwrap_optional(v))):
            injected_args[k] = value if is_maybe(v) else value.unwrap()
        elif is_maybe(v):
            injected_args[k] = Nothing
        elif is_optional(v):
            injected_args[k] = None

    kwargs.update(injected_args)
    return func(*args, **kwargs)


def inject[T, **P](target: Callable[P, T] | type[T]) -> Callable[P, T]:
    hints = get_type_hints(target)
    hints.pop('return', None)

    @wraps(target)
    def decorator(*args: P.args, **kwargs: P.kwargs) -> T:
        container = getattr(target, '__container__', root)
        return _invoke_target(target, hints, container, *args, **kwargs)

    _collector.add_target(target)
    return decorator



from abc import ABC, abstractmethod
from collections.abc import Awaitable, Mapping
from typing import override

from greyhorse.app.contexts import Context
from greyhorse.app.registries import MutDictRegistry
from greyhorse.maybe import Maybe
from greyhorse.utils.types import TypeWrapper


class Engine(ABC):
    def __init__(self, name: str) -> None:
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    @property
    @abstractmethod
    def active(self) -> bool: ...

    @abstractmethod
    def get_context[T: Context](self, kind: type[T]) -> Maybe[T]: ...

    @abstractmethod
    def start(self) -> Awaitable[None] | None: ...

    @abstractmethod
    def stop(self) -> Awaitable[None] | None: ...


class EngineReader[T: Engine](TypeWrapper[T], ABC):
    @abstractmethod
    def get_engine(self, name: str) -> Maybe[T]: ...


class EngineFactory[T: Engine](EngineReader[T], ABC):
    @abstractmethod
    def create_engine[T](self, name: str, config: T) -> Engine: ...

    @abstractmethod
    def destroy_engine(self, name: str) -> bool: ...

    @abstractmethod
    def get_engine_names(self) -> list[str]: ...

    @abstractmethod
    def get_engines(self, names: list[str] | None = None) -> Mapping[str, Engine]: ...


class SimpleEngineFactory[T: Engine](EngineFactory[T], ABC):
    def __init__(self) -> None:
        self._engines = MutDictRegistry[str, Engine]()

    @override
    def destroy_engine(self, name: str) -> bool:
        return self._engines.remove(name)

    @override
    def get_engine_names(self) -> list[str]:
        return [k for k, _ in self._engines.items()]

    @override
    def get_engine(self, name: str) -> Maybe[T]:
        return self._engines.get(name)

    @override
    def get_engines(self, names: list[str] | None = None) -> Mapping[str, T]:
        engines = {}

        if not names:
            names = self.get_engine_names()

        for name in names:
            if engine := self.get_engine(name).unwrap_or_none():
                engines[name] = engine

        return engines



from abc import ABC, abstractmethod

from greyhorse.error import Error, ErrorCase
from greyhorse.maybe import Maybe
from greyhorse.result import Result

from .operators import Operator
from .providers import Provider
from .visitor import Visitor


class ModuleError(Error):
    namespace = 'greyhorse.app.module'

    Component = ErrorCase(
        msg='{path}: Component error in module, details: "{details}"', path=str, details=str
    )

    Resource = ErrorCase(
        msg='{path}: Resource error in module: "{details}"', path=str, details=str
    )


class Module(ABC):
    __slots__ = ('_path',)

    def __init__(self, path: str) -> None:
        self._path = path

    @property
    def path(self) -> str:
        return self._path

    @abstractmethod
    def accept_visitor(self, visitor: Visitor) -> None: ...

    @abstractmethod
    def get_provider[P: Provider](self, prov_type: type[P]) -> Maybe[P]: ...

    @abstractmethod
    def add_provider[T](self, prov_type: type[Provider[T]], provider: Provider[T]) -> bool: ...

    @abstractmethod
    def remove_provider[T](self, prov_type: type[Provider[T]]) -> bool: ...

    @abstractmethod
    def add_resource[T](self, res_type: type[T], resource: T) -> bool: ...

    @abstractmethod
    def remove_resource[T](self, res_type: type[T]) -> bool: ...

    @abstractmethod
    def add_operator[T](self, operator: Operator[T]) -> bool: ...

    @abstractmethod
    def remove_operator[T](self, operator: Operator[T]) -> bool: ...

    @abstractmethod
    def create(self) -> Result[None, ModuleError]: ...

    @abstractmethod
    def setup(self) -> Result[None, ModuleError]: ...

    @abstractmethod
    def teardown(self) -> Result[None, ModuleError]: ...

    @abstractmethod
    def destroy(self) -> Result[None, ModuleError]: ...



from greyhorse.app.abc.controllers import ControllerError
from greyhorse.app.abc.operators import Operator
from greyhorse.app.entities.controllers import ResConf, ResourceController
from greyhorse.app.resources import make_container
from greyhorse.maybe import Just, Maybe


def test_res_controller() -> None:
    confs = [
        ResConf(type=int, required=True),
        ResConf(type=str, required=False),
        ResConf(type=Maybe, required=True),
    ]

    instance = ResourceController(confs)
    container = make_container()

    res = instance.setup(container)
    assert res.is_err()
    assert res.unwrap_err() == ControllerError.NoSuchResource(name='int')

    res = instance.teardown(container)
    assert res.is_err()
    assert res.unwrap_err() == ControllerError.NoSuchResource(name='int')

    operators = []
    instance.inspect(lambda member: operators.append(member.method()))

    int_op: Operator[int] = operators[0]
    str_op: Operator[str] = operators[1]
    maybe_op: Operator[Maybe] = operators[2]

    assert len(container.registry) == 0

    with container.context:
        assert int_op.accept(123)
        assert instance.setup(container).unwrap_err() == ControllerError.NoSuchResource(
            name='Maybe'
        )
        assert len(container.registry) == 1

    with container.context:
        assert str_op.accept('123')
        assert instance.setup(container).unwrap_err() == ControllerError.NoSuchResource(
            name='Maybe'
        )
        assert len(container.registry) == 2

    with container.context:
        assert maybe_op.accept(Just(123))
        assert instance.setup(container).is_ok()
        assert len(container.registry) == 3

    with container.context:
        assert instance.setup(container).is_ok()
        assert int_op.revoke().unwrap() == 123
        assert instance.teardown(container).unwrap_err() == ControllerError.NoSuchResource(
            name='int'
        )
        assert len(container.registry) == 3

    with container.context:
        assert int_op.accept(123)
        assert instance.setup(container).is_ok()
        assert str_op.revoke().unwrap() == '123'
        assert instance.teardown(container).is_ok()
        assert len(container.registry) == 1

    assert maybe_op.revoke().unwrap() == Just(123)



from typing import Any, override

from greyhorse.logging import logger
from greyhorse.maybe import Maybe, Nothing
from greyhorse.result import Ok, Result

from ..abc.module import Module, ModuleError
from ..abc.operators import Operator
from ..abc.providers import Provider
from ..abc.visitor import Visitor
from ..entities.components import Component
from ..registries import MutDictRegistry
from ..resources.manager import ResourceManager
from ..schemas.components import ModuleConf


class SyncModule(Module):
    def __init__(self, path: str, conf: ModuleConf, components: list[Component]) -> None:
        super().__init__(path)
        self._conf = conf
        self._rm = ResourceManager()

        self._operators: list[Operator] = []
        self._resources = MutDictRegistry[type, Any]()
        self._providers = MutDictRegistry[type[Provider], Provider]()
        self._components: dict[str, Component] = {c.name: c for c in components}

    @property
    def conf(self) -> ModuleConf:
        return self._conf

    @override
    def accept_visitor(self, visitor: Visitor) -> None:
        visitor.start_module(self)
        for comp in self._components.values():
            comp.accept_visitor(visitor)
        visitor.finish_module(self)

    @override
    def get_provider[P: Provider](self, prov_type: type[P]) -> Maybe[P]:
        if prov_type in self._conf.providers:
            return self._providers.get(prov_type)
        return Nothing

    @override
    def add_provider[T](self, prov_type: type[Provider[T]], provider: Provider[T]) -> bool:
        if prov_type in self._conf.provider_claims:
            return self._providers.add(prov_type, provider)
        return False

    @override
    def remove_provider[T](self, prov_type: type[Provider[T]]) -> bool:
        if prov_type in self._conf.provider_claims:
            return self._providers.remove(prov_type)
        return False

    @override
    def add_resource[T](self, res_type: type[T], resource: T) -> bool:
        if res_type in self._conf.resource_claims:
            return self._resources.add(res_type, resource)
        return False

    @override
    def remove_resource[T](self, res_type: type[T]) -> bool:
        if res_type in self._conf.resource_claims:
            return self._resources.remove(res_type)
        return False

    @override
    def add_operator[T](self, operator: Operator[T]) -> bool:
        for res_type in self._conf.operators:
            if issubclass(res_type, operator.wrapped_type):
                self._operators.append(operator)
                return True
        return False

    @override
    def remove_operator[T](self, operator: Operator[T]) -> bool:
        for res_type in self._conf.operators:
            if issubclass(res_type, operator.wrapped_type):
                self._operators.remove(operator)
                return True
        return False

    @override
    def create(self) -> Result[None, ModuleError]:
        logger.info('{path}: Module create'.format(path=self._path))

        for component in self._components.values():
            if not (
                res := component.create().map_err(
                    lambda e: ModuleError.Component(path=self._path, details=e.message)
                )
            ):
                return res

        logger.info('{path}: Module create successful'.format(path=self._path))
        return Ok()

    @override
    def setup(self) -> Result[None, ModuleError]:
        logger.info('{path}: Module setup'.format(path=self._path))

        for component in self._components.values():
            comp_conf = self._conf.components[component.name]

            for res_type in comp_conf.resource_claims:
                if res := self._resources.get(res_type).unwrap_or_none():
                    component.add_resource(res_type, res)

            for res_type in comp_conf.operators:
                for op in component.get_operators(res_type):  # type: ignore
                    if not (
                        res := self._rm.setup_operator(op, providers=self._providers).map_err(
                            lambda e: ModuleError.Resource(path=self._path, details=e.message)
                        )
                    ):
                        return res  # type: ignore

            if not (
                res := component.setup().map_err(
                    lambda e: ModuleError.Component(path=self._path, details=e.message)
                )
            ):
                return res

            for prov_type in comp_conf.providers:
                if prov := component.get_provider(prov_type).unwrap_or_none():
                    self._providers.add(prov_type, prov)

        for op in self._operators:
            if not (
                res := self._rm.setup_operator(op, providers=self._providers).map_err(
                    lambda e: ModuleError.Resource(path=self._path, details=e.message)
                )
            ):
                return res  # type: ignore

        logger.info('{path}: Module setup successful'.format(path=self._path))
        return Ok()

    @override
    def teardown(self) -> Result[None, ModuleError]:
        logger.info('{path}: Module teardown'.format(path=self._path))

        for op in reversed(self._operators):
            if not (
                res := self._rm.teardown_operator(op).map_err(
                    lambda e: ModuleError.Resource(path=self._path, details=e.message)
                )
            ):
                return res  # type: ignore

        for component in reversed(self._components.values()):
            comp_conf = self._conf.components[component.name]

            for prov_type in reversed(comp_conf.providers):
                self._providers.remove(prov_type)

            if not (
                res := component.teardown().map_err(
                    lambda e: ModuleError.Component(path=self._path, details=e.message)
                )
            ):
                return res

            for res_type in reversed(comp_conf.operators):
                for op in component.get_operators(res_type):  # type: ignore
                    if not (
                        res := self._rm.teardown_operator(op).map_err(
                            lambda e: ModuleError.Resource(path=self._path, details=e.message)
                        )
                    ):
                        return res  # type: ignore

            for res_type in reversed(comp_conf.resource_claims):
                if self._resources.has(res_type):
                    component.remove_resource(res_type)

        logger.info('{path}: Module teardown successful'.format(path=self._path))
        return Ok()

    @override
    def destroy(self) -> Result[None, ModuleError]:
        logger.info('{path}: Module teardown'.format(path=self._path))

        for component in reversed(self._components.values()):
            if not (
                res := component.destroy().map_err(
                    lambda e: ModuleError.Component(path=self._path, details=e.message)
                )
            ):
                return res

        if not (
            res := self._rm.teardown().map_err(
                lambda e: ModuleError.Resource(path=self._path, details=e.message)
            )
        ):
            return res

        logger.info('{path}: Module teardown successful'.format(path=self._path))
        return Ok()



from typing import Any

from greyhorse.logging import logger
from greyhorse.utils.imports import import_path
from greyhorse.utils.injectors import ParamsInjector
from greyhorse.utils.invoke import invoke_sync
from greyhorse.utils.project import get_project_path

from ..entities.application import Application


def load_packages(app: Application, key: str, **kwargs: dict[str, Any]) -> None:
    from tomlkit import parse

    packages = {}
    injector = ParamsInjector()
    pyproject_toml_path = get_project_path()
    pyproject_toml = parse(string=pyproject_toml_path.read_text())

    if project := pyproject_toml.get('project'):
        if key:
            if section := project.get(key):
                packages = section.get('packages', dict())
        else:
            packages = project.get('packages', dict())

    key = key or 'package'

    for name, dotted_path in packages.items():
        logger.info(f'Import package: "{name}"')
        dotted_path = dotted_path if ':' in dotted_path else f'{dotted_path}:{key}_init'

        if entrypoint := import_path(dotted_path):
            if not callable(entrypoint):
                continue
            logger.info(f'Application "{app.name}": load package "{name}"')
            injected_args = injector(entrypoint, values={'app': app, **kwargs})
            invoke_sync(entrypoint, *injected_args.args, **injected_args.kwargs)



import inspect
from asyncio import iscoroutinefunction
from functools import partial
from types import EllipsisType, NoneType, UnionType, new_class
from typing import Any, TypeVar, TypeVarTuple, Union, get_args, get_origin

from .strings import capitalize

_TYPES_CACHE = {}


class TypeWrapper[T]:
    __wrapped_type__: type | tuple[type, ...] = type

    @property
    def wrapped_type(self) -> type | tuple[type, ...]:
        return self.__wrapped_type__

    @classmethod
    def __generate_typename__(
        cls,
        types: type | TypeVar | TypeVarTuple | tuple[type | TypeVar | TypeVarTuple, ...],
        include_base_name: bool = True,
    ) -> str:
        if isinstance(types, tuple):
            type_args = [cls.__generate_typename__(a, False) for a in types]
            type_name = f'{''.join(type_args)}'
        elif isinstance(types, TypeVar):
            type_name = '~'
        elif isinstance(types, NoneType):
            type_name = 'None'
        elif isinstance(types, UnionType):
            type_args = [cls.__generate_typename__(a, False) for a in types.__args__]
            type_name = f'{''.join(type_args)}Union'
        elif isinstance(types, EllipsisType):
            type_name = '...'
        elif hasattr(types, '__args__'):
            type_args = [cls.__generate_typename__(a, False) for a in types.__args__]
            type_name = f'{''.join(type_args)}{capitalize(types.__name__)}'
        else:
            type_name = f'{capitalize(types.__name__)}'

        if include_base_name:
            if cls.__name__.startswith('_'):
                type_name = f'_{type_name}{cls.__name__[1:]}'
            else:
                type_name = f'{type_name}{cls.__name__}'

        return type_name

    def __init_subclass__(cls, **kwargs: dict[str, Any]) -> None:
        if spec := kwargs.pop('spec', None):
            if not isinstance(spec, list):
                spec = [spec]

            for spec_class in spec:  # type: type
                type_name = cls.__generate_typename__(spec_class, False)
                _TYPES_CACHE[type_name] = cls

        super().__init_subclass__(**kwargs)

    def __class_getitem__(
        cls, types: type | TypeVar | TypeVarTuple | tuple[type | TypeVar | TypeVarTuple, ...]
    ) -> type:
        if isinstance(types, TypeVar | TypeVarTuple):
            # noinspection PyUnresolvedReferences
            return super().__class_getitem__(types)

        type_name = cls.__generate_typename__(types)

        if class_ := _TYPES_CACHE.get(type_name):
            return class_

        bases = [cls]

        if not isinstance(types, tuple) and hasattr(types, '__base__'):
            base_type_name = '~' + cls.__generate_typename__(types.__base__)
            if base_class := _TYPES_CACHE.get(base_type_name):
                bases = [base_class, *bases]

        if hasattr(cls, '__slots__'):
            cls.__slots__ = (*cls.__slots__, '__wrapped_type__')
            attrs = {k: v for k, v in cls.__dict__.items() if k in cls.__slots__}
            attrs['__wrapped_type__'] = types
            attrs['__module__'] = cls.__module__
            class_ = new_class(type_name, tuple(bases), exec_body=lambda d: d.update(attrs))

        else:
            attrs = dict(cls.__dict__)
            attrs.pop('__class_getitem__', None)
            attrs['__wrapped_type__'] = types
            attrs['__module__'] = cls.__module__

            class_ = new_class(type_name, tuple(bases), exec_body=lambda d: d.update(attrs))

        _TYPES_CACHE[type_name] = class_
        return class_


def is_optional[T](value: type[T]) -> bool:
    if get_origin(value) is Union or isinstance(value, UnionType):
        type_args = get_args(value)
        if len(type_args) > 1:
            return type_args[1] is NoneType
    return False


def is_maybe[T](value: type[T]) -> bool:
    from greyhorse.maybe import Maybe

    return get_origin(value) is Maybe or isinstance(value, Maybe)


def is_awaitable(f: object) -> bool:
    while isinstance(f, partial):
        f = f.func
    f = inspect.unwrap(f)
    return iscoroutinefunction(f) or inspect.isawaitable(f)


def unwrap_optional[T](value: type[T]) -> type[T]:
    if get_origin(value) is Union or isinstance(value, UnionType):
        type_args = get_args(value)
        if len(type_args) >= 1:
            return type_args[0]
    return value


def unwrap_maybe[T](value: type[T]) -> type[T]:
    from greyhorse.maybe import Maybe

    if get_origin(value) is Maybe or isinstance(value, Maybe):
        type_args = get_args(value)
        if len(type_args) >= 1:
            return type_args[0]
    return value



from typing import Any

from pydantic import BaseModel, Field, PrivateAttr

from greyhorse.utils.invoke import caller_path

from ..abc.controllers import ControllerFactories
from ..abc.providers import (
    FactoryProvider,
    ForwardProvider,
    MutProvider,
    Provider,
    SharedProvider,
)
from ..abc.services import ServiceFactories
from ..schemas.elements import CtrlConf, SvcConf


class ComponentConf(BaseModel):
    enabled: bool = Field(default=True)

    resource_claims: list[type] = Field(default_factory=list)
    operators: list[type] = Field(default_factory=list)
    providers: list[type[Provider]] = Field(default_factory=list)

    controllers: list[CtrlConf] = Field(default_factory=list)
    services: list[SvcConf] = Field(default_factory=list)

    controller_factories: ControllerFactories = Field(default_factory=dict)
    service_factories: ServiceFactories = Field(default_factory=dict)


class ModuleConf(BaseModel):
    enabled: bool = Field(default=True)

    provider_claims: list[type[ForwardProvider] | type[SharedProvider]] = Field(
        default_factory=list
    )
    resource_claims: list[type] = Field(default_factory=list)

    operators: list[type] = Field(default_factory=list)
    providers: list[type[FactoryProvider] | type[MutProvider]] = Field(default_factory=list)

    components: dict[str, ComponentConf] = Field(default_factory=dict)


class ModuleComponentConf(ComponentConf):
    path: str = Field(frozen=True)
    args: dict[str, Any] = Field(default_factory=dict, frozen=True)
    _init_path: list[str] = PrivateAttr(default_factory=lambda: caller_path(5))
    _conf: ModuleConf | None = PrivateAttr(default=None)



import contextlib
from abc import ABC, abstractmethod
from collections.abc import AsyncGenerator, Awaitable, Callable, Generator
from typing import Any, Self, override

from greyhorse.maybe import Just, Maybe, Nothing
from greyhorse.utils.types import TypeWrapper


class TypeFactory[T](TypeWrapper[T], ABC):
    scoped: bool = False
    cache: bool = False

    @abstractmethod
    def create(self, key: type[T]) -> Maybe[T] | Awaitable[Maybe[T]]: ...

    def destroy(self, instance: T) -> None | Awaitable[None]: ...

    def has(self, key: type[T]) -> bool:
        if Any is self.wrapped_type:
            return True
        return issubclass(self.wrapped_type, key)

    @classmethod
    def from_fn(cls, fn: Callable[[T], T]) -> Self:
        return _FnTypeFactory[cls.__wrapped_type__](fn)

    @classmethod
    def from_class(cls, cls_: type[T]) -> Self:
        return _ClassTypeFactory[cls.__wrapped_type__](cls_)

    @classmethod
    def from_instance(cls, instance: T) -> Self:
        return _InstanceTypeFactory[cls.__wrapped_type__](instance)

    @classmethod
    def from_syncgen(cls, gen_fn: Callable[[], Generator[T, T, None]]) -> Self:
        return _SyncGenTypeFactory[cls.__wrapped_type__](gen_fn)

    @classmethod
    def from_asyncgen(cls, gen_fn: Callable[[], AsyncGenerator[T, T]]) -> Self:
        return _AsyncGenTypeFactory[cls.__wrapped_type__](gen_fn)


class _FnTypeFactory[T](TypeFactory[T]):
    __slots__ = ('_fn',)

    def __init__(self, fn: Callable[[T], T]) -> None:
        self._fn = fn

    @override
    def create(self, key: type[T]) -> Maybe[T]:
        return Maybe(self._fn(key))


class _ClassTypeFactory[T](TypeFactory[T]):
    __slots__ = ('_cls',)

    def __init__(self, cls: type[T]) -> None:
        self._cls = cls

    @override
    def create(self, _: type[T]) -> Maybe[T]:
        return Just(self._cls())


class _InstanceTypeFactory[T](TypeFactory[T]):
    __slots__ = ('_instance',)

    cache = True

    def __init__(self, instance: T) -> None:
        self._instance = instance

    @override
    def create(self, _: type[T]) -> Maybe[T]:
        return Just(self._instance)


class _SyncGenTypeFactory[T](TypeFactory[T]):
    __slots__ = ('_gen_fn', '_gens')

    scoped = True

    def __init__(self, generator_fn: Callable[[], Generator[T, T, None]]) -> None:
        self._gen_fn = generator_fn
        self._gens: dict[int, Generator[T, T, None]] = {}

    @override
    def create(self, _: type[T]) -> Maybe[T]:
        gen = self._gen_fn()

        try:
            instance = next(gen)

        except StopIteration as e:
            instance = e.value

        if instance is None:
            return Nothing

        self._gens[id(instance)] = gen
        return Just(instance)

    @override
    def destroy(self, instance: T) -> None:
        key = id(instance)
        if not (gen := self._gens.pop(key, None)):
            return

        with contextlib.suppress(StopIteration):
            gen.send(instance)


class _AsyncGenTypeFactory[T](TypeFactory[T]):
    __slots__ = ('_gen_fn', '_gens')

    scoped = True

    def __init__(self, generator_fn: Callable[[], AsyncGenerator[T, T]]) -> None:
        self._gen_fn = generator_fn
        self._gens: dict[int, AsyncGenerator[T, T]] = {}

    @override
    async def create(self, _: type[T]) -> Maybe[T]:
        gen = self._gen_fn()

        try:
            instance = await anext(gen)

        except StopAsyncIteration:
            instance = None

        if instance is None:
            return Nothing

        self._gens[id(instance)] = gen
        return Just(instance)

    @override
    async def destroy(self, instance: T) -> None:
        key = id(instance)
        if not (gen := self._gens.pop(key, None)):
            return

        with contextlib.suppress(StopAsyncIteration):
            await gen.asend(instance)



from collections.abc import Callable
from functools import partial
from importlib import import_module


def import_path(dotted_path: str, package: str | None = None) -> object:
    if ':' in dotted_path:
        module_path, attr_path = dotted_path.rsplit(':', maxsplit=1)
    else:
        module_path, attr_path = dotted_path.rsplit('.', maxsplit=1)

    module = import_module(module_path, package)

    if ':' in dotted_path:
        result = module
        for attr_name in attr_path.rsplit('.'):
            result = getattr(result, attr_name)
        return result

    return getattr(module, attr_path)


def lazy_import(
    dotted_path: str, callable: bool = False, as_partial: bool = False
) -> Callable[[], object]:
    if callable:

        def inner(*args, **kwargs) -> object:  # noqa: ANN002,ANN003
            obj = import_path(dotted_path)
            if as_partial:
                return partial(obj, *args, **kwargs)
            return obj(*args, **kwargs)

    else:

        def inner() -> object:
            return import_path(dotted_path)

    return inner



import contextlib
from collections.abc import AsyncGenerator, Callable, Generator
from copy import deepcopy
from typing import Any, override

from greyhorse.maybe import Maybe, Nothing
from greyhorse.result import Err, Ok, Result

from .abc.operators import Operator
from .abc.providers import (
    BorrowError,
    BorrowMutError,
    FactoryError,
    FactoryProvider,
    ForwardError,
    ForwardProvider,
    MutProvider,
    SharedProvider,
)
from .contexts import AsyncContext, AsyncMutContext, ContextBuilder, SyncContext, SyncMutContext


class _BasicRefBox:
    __slots__ = ('_shared_counter', '_acq_counter')

    allow_borrow_when_acquired = False
    allow_acq_when_borrowed = False
    allow_multiple_acquisition = False

    def __init__(self) -> None:
        self._shared_counter = 0
        self._acq_counter = 0

    def _borrow[T](self, value: T) -> Result[T, BorrowError]:
        if not self.allow_borrow_when_acquired and self._acq_counter > 0:
            return BorrowError.BorrowedAsMutable(name=type(value).__name__).to_result()
        self._shared_counter += 1
        return Ok(value)

    def _borrow_mut[T](self, value: T) -> Result[T, BorrowMutError]:
        if not self.allow_multiple_acquisition and self._acq_counter > 0:
            return BorrowMutError.AlreadyBorrowed(name=type(value).__name__).to_result()
        if not self.allow_acq_when_borrowed and self._shared_counter > 0:
            return BorrowMutError.BorrowedAsImmutable(name=type(value).__name__).to_result()
        self._acq_counter += 1
        return Ok(value)

    @staticmethod
    def _ensure_maybe(value: object) -> Maybe:
        if not isinstance(value, Maybe):
            value = Maybe(value)
        return value


class SharedRefBox[T](_BasicRefBox, SharedProvider[T]):
    __slots__ = ('_factory', '_copy_maker')

    def __init__(
        self, factory: Callable[[], Maybe[T] | T], copy_maker: Callable[[T], T] = deepcopy
    ) -> None:
        super().__init__()
        self._factory = factory
        self._copy_maker = copy_maker

    @override
    def borrow(self) -> Result[T, BorrowError]:
        return (
            self._ensure_maybe(self._factory())
            .map_or_else(
                lambda: BorrowError.Empty(name=self.wrapped_type.__name__).to_result(),
                self._borrow,
            )
            .map(self._copy_maker)
        )

    @override
    def reclaim(self, instance: T) -> None:
        self._shared_counter -= 1
        del instance


class MutRefBox[T](_BasicRefBox, MutProvider[T]):
    __slots__ = ('_factory', '_copy_maker')

    def __init__(
        self, factory: Callable[[], Maybe[T] | T], copy_maker: Callable[[T], T] = deepcopy
    ) -> None:
        super().__init__()
        self._factory = factory
        self._copy_maker = copy_maker

    @override
    def acquire(self) -> Result[T, BorrowMutError]:
        return (
            self._ensure_maybe(self._factory())
            .map_or_else(
                lambda: BorrowError.Empty(name=self.wrapped_type.__name__).to_result(),
                self._borrow_mut,
            )
            .map(self._copy_maker)
        )

    @override
    def release(self, instance: T) -> None:
        self._acq_counter -= 1
        del instance


class OwnerRefBox[TS, TM](_BasicRefBox, SharedProvider[TS], MutProvider[TM]):
    __slots__ = ('_factory', '_mut_factory', '_copy_maker', '_mut_copy_maker')

    def __init__(
        self,
        factory: Callable[[], Maybe[TS] | TS],
        mut_factory: Callable[[], Maybe[TM] | TM],
        copy_maker: Callable[[TS], TS] = deepcopy,
        mut_copy_maker: Callable[[TM], TM] = deepcopy,
    ) -> None:
        super().__init__()
        self._factory = factory
        self._mut_factory = mut_factory
        self._copy_maker = copy_maker
        self._mut_copy_maker = mut_copy_maker

    @override
    def borrow(self) -> Result[TS, BorrowError]:
        return (
            self._ensure_maybe(self._factory())
            .map_or_else(
                lambda: BorrowError.Empty(name=self.wrapped_type[0].__name__).to_result(),
                self._borrow,
            )
            .map(self._copy_maker)
        )

    @override
    def reclaim(self, instance: TS) -> None:
        self._shared_counter -= 1
        del instance

    @override
    def acquire(self) -> Result[TM, BorrowMutError]:
        return (
            self._ensure_maybe(self._mut_factory())
            .map_or_else(
                lambda: BorrowError.Empty(name=self.wrapped_type[1].__name__).to_result(),
                self._borrow_mut,
            )
            .map(self._mut_copy_maker)
        )

    @override
    def release(self, instance: TM) -> None:
        self._acq_counter -= 1
        del instance


class SharedCtxRefBox[T](_BasicRefBox, SharedProvider[T]):
    __slots__ = ('_kind', '_params', '_factory')

    def __init__(
        self,
        kind: type[SyncContext[T] | AsyncContext[T]],
        factory: Callable[[], T],
        **params: dict[str, Any],
    ) -> None:
        super().__init__()
        self._kind = kind
        self._params = params
        self._factory = factory

    @override
    def borrow(self) -> Result[SyncContext[T] | AsyncContext[T], BorrowError]:
        if not self.allow_borrow_when_acquired and self._acq_counter > 0:
            return BorrowError.BorrowedAsMutable(name=self.wrapped_type.__name__).to_result()
        self._shared_counter += 1

        ctx_builder = ContextBuilder[self._kind, self.wrapped_type](
            self._factory, **self._params
        )
        return Ok(ctx_builder.build())

    @override
    def reclaim(self, instance: SyncContext[T] | AsyncContext[T]) -> None:
        self._shared_counter -= 1
        del instance


class MutCtxRefBox[T](_BasicRefBox, MutProvider[T]):
    __slots__ = ('_mut_kind', '_params', '_factory')

    def __init__(
        self,
        mut_kind: type[SyncMutContext[T] | AsyncMutContext[T]],
        factory: Callable[[], T],
        **params: dict[str, Any],
    ) -> None:
        super().__init__()
        self._mut_kind = mut_kind
        self._params = params
        self._factory = factory

    @override
    def acquire(self) -> Result[SyncMutContext[T] | AsyncMutContext[T], BorrowMutError]:
        if not self.allow_multiple_acquisition and self._acq_counter > 0:
            return BorrowMutError.AlreadyBorrowed(name=self.wrapped_type.__name__).to_result()
        if not self.allow_acq_when_borrowed and self._shared_counter > 0:
            return BorrowMutError.BorrowedAsImmutable(
                name=self.wrapped_type.__name__
            ).to_result()
        self._acq_counter += 1

        ctx_builder = ContextBuilder[self._mut_kind, self.wrapped_type](
            self._factory, **self._params
        )
        return Ok(ctx_builder.build())

    @override
    def release(self, instance: SyncMutContext[T] | AsyncMutContext[T]) -> None:
        self._acq_counter -= 1
        del instance


class OwnerCtxRefBox[TS, TM](_BasicRefBox, SharedProvider[TS], MutProvider[TM]):
    __slots__ = ('_kind', '_mut_kind', '_params', '_mut_params', '_factory', '_mut_factory')

    def __init__(
        self,
        kind: type[SyncContext[TS] | AsyncContext[TS]],
        mut_kind: type[SyncMutContext[TM] | AsyncMutContext[TM]],
        factory: Callable[[], TS],
        mut_factory: Callable[[], TM],
        params: dict[str, Any] | None = None,
        mut_params: dict[str, Any] | None = None,
    ) -> None:
        super().__init__()
        self._kind = kind
        self._mut_kind = mut_kind
        self._params = params or {}
        self._mut_params = mut_params or {}
        self._factory = factory
        self._mut_factory = mut_factory

    @override
    def borrow(self) -> Result[SyncContext[TS] | AsyncContext[TS], BorrowError]:
        if not self.allow_borrow_when_acquired and self._acq_counter > 0:
            return BorrowError.BorrowedAsMutable(name=self.wrapped_type[0].__name__).to_result()
        self._shared_counter += 1

        ctx_builder = ContextBuilder[self._kind, self.wrapped_type[0]](
            self._factory, **self._params
        )
        return Ok(ctx_builder.build())

    @override
    def reclaim(self, instance: SyncContext[TS] | AsyncContext[TS]) -> None:
        self._shared_counter -= 1
        del instance

    @override
    def acquire(self) -> Result[SyncMutContext[TM] | AsyncMutContext[TM], BorrowMutError]:
        if not self.allow_multiple_acquisition and self._acq_counter > 0:
            return BorrowMutError.AlreadyBorrowed(
                name=self.wrapped_type[1].__name__
            ).to_result()
        if not self.allow_acq_when_borrowed and self._shared_counter > 0:
            return BorrowMutError.BorrowedAsImmutable(
                name=self.wrapped_type[1].__name__
            ).to_result()
        self._acq_counter += 1

        ctx_builder = ContextBuilder[self._mut_kind, self.wrapped_type[1]](
            self._mut_factory, **self._mut_params
        )
        return Ok(ctx_builder.build())

    @override
    def release(self, instance: SyncMutContext[TM] | AsyncMutContext[TM]) -> None:
        self._acq_counter -= 1
        del instance


class ForwardBox[T](Operator[T], ForwardProvider[T]):
    __slots__ = ('_value',)

    def __init__(self, value: T | None = None) -> None:
        self._value = Maybe(value)

    @override
    def accept(self, value: T) -> bool:
        if self._value.is_just():
            return False
        self._value = Maybe(value)
        return True

    @override
    def revoke(self) -> Maybe[T]:
        value, self._value = self._value, Nothing
        return value

    @override
    def take(self) -> Result[T, ForwardError]:
        value, self._value = self._value, Nothing
        return value.map(Ok).unwrap_or(
            ForwardError.Empty(name=self.wrapped_type.__name__).to_result()
        )

    @override
    def drop(self, instance: T) -> None:
        del instance

    @override
    def __bool__(self) -> bool:
        return self._value.is_just()


class PermanentForwardBox[T](ForwardBox[T]):
    @override
    def take(self) -> Result[T, ForwardError]:
        return self._value.map(Ok).unwrap_or(
            ForwardError.Empty(name=self.wrapped_type.__name__).to_result()
        )

    @override
    def drop(self, instance: T) -> None:
        pass


class SyncSharedGenBox[T](SharedProvider[T]):
    __slots__ = ('_gen_fn', '_gens')

    def __init__(self, generator_fn: Callable[[], Generator[T, T, None]]) -> None:
        self._gen_fn = generator_fn
        self._gens: dict[int, Generator[T, T, None]] = {}

    @override
    def borrow(self) -> Result[T, BorrowError]:
        try:
            gen = self._gen_fn()

        except Exception as e:
            return BorrowError.Unexpected(
                name=self.wrapped_type.__name__, details=str(e)
            ).to_result()

        try:
            instance = next(gen)

        except StopIteration as e:
            instance = e.value

        match instance:
            case Ok(instance):
                pass
            case Err(e):
                if isinstance(e, BorrowError):
                    return e.to_result()
                return BorrowError.Unexpected(
                    name=self.wrapped_type.__name__, details=e.message
                ).to_result()
            case None:
                return BorrowError.InsufficientDeps(name=self.wrapped_type.__name__).to_result()
            case instance:
                pass

        self._gens[id(instance)] = gen
        return Ok(instance)

    @override
    def reclaim(self, instance: T) -> None:
        key = id(instance)
        if not (gen := self._gens.pop(key, None)):
            return

        with contextlib.suppress(StopIteration):
            gen.send(instance)


class AsyncSharedGenBox[T](SharedProvider[T]):
    __slots__ = ('_gen_fn', '_gens')

    def __init__(self, generator_fn: Callable[[], AsyncGenerator[T, T]]) -> None:
        self._gen_fn = generator_fn
        self._gens: dict[int, AsyncGenerator[T, T]] = {}

    @override
    async def borrow(self) -> Result[T, BorrowError]:
        try:
            gen = self._gen_fn()

        except Exception as e:
            return BorrowError.Unexpected(
                name=self.wrapped_type.__name__, details=str(e)
            ).to_result()

        try:
            instance = await anext(gen)

        except StopAsyncIteration:
            instance = None

        match instance:
            case Ok(instance):
                pass
            case Err(e):
                if isinstance(e, BorrowError):
                    return e.to_result()
                return BorrowError.Unexpected(
                    name=self.wrapped_type.__name__, details=e.message
                ).to_result()
            case None:
                return BorrowError.InsufficientDeps(name=self.wrapped_type.__name__).to_result()
            case instance:
                pass

        self._gens[id(instance)] = gen
        return Ok(instance)

    @override
    async def reclaim(self, instance: T) -> None:
        key = id(instance)
        if not (gen := self._gens.pop(key, None)):
            return

        with contextlib.suppress(StopAsyncIteration):
            await gen.asend(instance)


class SyncMutGenBox[T](MutProvider[T]):
    __slots__ = ('_gen_fn', '_gens')

    def __init__(self, generator_fn: Callable[[], Generator[T, T, None]]) -> None:
        self._gen_fn = generator_fn
        self._gens: dict[int, Generator[T, T, None]] = {}

    @override
    def acquire(self) -> Result[T, BorrowMutError]:
        try:
            gen = self._gen_fn()

        except Exception as e:
            return BorrowMutError.Unexpected(
                name=self.wrapped_type.__name__, details=str(e)
            ).to_result()

        try:
            instance = next(gen)

        except StopIteration as e:
            instance = e.value

        match instance:
            case Ok(instance):
                pass
            case Err(e):
                if isinstance(e, BorrowMutError):
                    return e.to_result()
                return BorrowMutError.Unexpected(
                    name=self.wrapped_type.__name__, details=e.message
                ).to_result()
            case None:
                return BorrowMutError.InsufficientDeps(
                    name=self.wrapped_type.__name__
                ).to_result()
            case instance:
                pass

        self._gens[id(instance)] = gen
        return Ok(instance)

    @override
    def release(self, instance: T) -> None:
        key = id(instance)
        if not (gen := self._gens.pop(key, None)):
            return

        with contextlib.suppress(StopIteration):
            gen.send(instance)


class AsyncMutGenBox[T](MutProvider[T]):
    __slots__ = ('_gen_fn', '_gens')

    def __init__(self, generator_fn: Callable[[], AsyncGenerator[T, T]]) -> None:
        self._gen_fn = generator_fn
        self._gens: dict[int, AsyncGenerator[T, T]] = {}

    @override
    async def acquire(self) -> Result[T, BorrowMutError]:
        try:
            gen = self._gen_fn()

        except Exception as e:
            return BorrowMutError.Unexpected(
                name=self.wrapped_type.__name__, details=str(e)
            ).to_result()

        try:
            instance = await anext(gen)

        except StopAsyncIteration:
            instance = None

        match instance:
            case Ok(instance):
                pass
            case Err(e):
                if isinstance(e, BorrowMutError):
                    return e.to_result()
                return BorrowMutError.Unexpected(
                    name=self.wrapped_type.__name__, details=e.message
                ).to_result()
            case None:
                return BorrowMutError.InsufficientDeps(
                    name=self.wrapped_type.__name__
                ).to_result()
            case instance:
                pass

        self._gens[id(instance)] = gen
        return Ok(instance)

    @override
    async def release(self, instance: T) -> None:
        key = id(instance)
        if not (gen := self._gens.pop(key, None)):
            return

        with contextlib.suppress(StopAsyncIteration):
            await gen.asend(instance)


class SyncFactoryGenBox[T](FactoryProvider[T]):
    __slots__ = ('_gen_fn', '_gens')

    def __init__(self, generator_fn: Callable[[], Generator[T, T, None]]) -> None:
        self._gen_fn = generator_fn
        self._gens: dict[int, Generator[T, T, None]] = {}

    @override
    def create(self) -> Result[T, FactoryError]:
        try:
            gen = self._gen_fn()

        except Exception as e:
            return FactoryError.Unexpected(
                name=self.wrapped_type.__name__, details=str(e)
            ).to_result()

        try:
            instance = next(gen)

        except StopIteration as e:
            instance = e.value

        match instance:
            case Ok(instance):
                pass
            case Err(e):
                if isinstance(e, FactoryError):
                    return e.to_result()
                return FactoryError.Unexpected(
                    name=self.wrapped_type.__name__, details=e.message
                ).to_result()
            case None:
                return FactoryError.InsufficientDeps(
                    name=self.wrapped_type.__name__
                ).to_result()
            case instance:
                pass

        self._gens[id(instance)] = gen
        return Ok(instance)

    @override
    def destroy(self, instance: T) -> None:
        key = id(instance)
        if not (gen := self._gens.pop(key, None)):
            return

        with contextlib.suppress(StopIteration):
            gen.send(instance)


class AsyncFactoryGenBox[T](FactoryProvider[T]):
    __slots__ = ('_gen_fn', '_gens')

    def __init__(self, generator_fn: Callable[[], AsyncGenerator[T, T]]) -> None:
        self._gen_fn = generator_fn
        self._gens: dict[int, AsyncGenerator[T, T]] = {}

    @override
    async def create(self) -> Result[T, FactoryError]:
        try:
            gen = self._gen_fn()

        except Exception as e:
            return FactoryError.Unexpected(
                name=self.wrapped_type.__name__, details=str(e)
            ).to_result()

        try:
            instance = await anext(gen)

        except StopAsyncIteration:
            instance = None

        match instance:
            case Ok(instance):
                pass
            case Err(e):
                if isinstance(e, FactoryError):
                    return e.to_result()
                return FactoryError.Unexpected(
                    name=self.wrapped_type.__name__, details=e.message
                ).to_result()
            case None:
                return FactoryError.InsufficientDeps(
                    name=self.wrapped_type.__name__
                ).to_result()
            case instance:
                pass

        self._gens[id(instance)] = gen
        return Ok(instance)

    @override
    async def destroy(self, instance: T) -> None:
        key = id(instance)
        if not (gen := self._gens.pop(key, None)):
            return

        with contextlib.suppress(StopAsyncIteration):
            await gen.asend(instance)


class SyncForwardGenBox[T](ForwardProvider[T]):
    __slots__ = ('_gen', '_moved_out')

    def __init__(self, generator: Generator[T, T, None]) -> None:
        self._gen = generator
        self._moved_out = False

    @override
    def take(self) -> Result[T, ForwardError]:
        if self._moved_out:
            return ForwardError.MovedOut(name=self.wrapped_type.__name__).to_result()

        try:
            instance = next(self._gen)

        except StopIteration as e:
            instance = e.value

        match instance:
            case Ok(instance):
                pass
            case Err(e):
                if isinstance(e, ForwardError):
                    return e.to_result()
                return ForwardError.Unexpected(
                    name=self.wrapped_type.__name__, details=e.message
                ).to_result()
            case None:
                return ForwardError.InsufficientDeps(
                    name=self.wrapped_type.__name__
                ).to_result()
            case instance:
                pass

        self._moved_out = True
        return Ok(instance)

    @override
    def drop(self, instance: T) -> None:
        if not self._moved_out:
            return

        with contextlib.suppress(StopIteration):
            self._gen.send(instance)

    @override
    def __bool__(self) -> bool:
        return not self._moved_out


class AsyncForwardGenBox[T](ForwardProvider[T]):
    __slots__ = ('_gen', '_moved_out')

    def __init__(self, generator: AsyncGenerator[T, T]) -> None:
        self._gen = generator
        self._moved_out = False

    @override
    async def take(self) -> Result[T, ForwardError]:
        if self._moved_out:
            return ForwardError.MovedOut(name=self.wrapped_type.__name__).to_result()

        try:
            instance = await anext(self._gen)

        except StopAsyncIteration:
            instance = None

        match instance:
            case Ok(instance):
                pass
            case Err(e):
                if isinstance(e, ForwardError):
                    return e.to_result()
                return ForwardError.Unexpected(
                    name=self.wrapped_type.__name__, details=e.message
                ).to_result()
            case None:
                return ForwardError.InsufficientDeps(
                    name=self.wrapped_type.__name__
                ).to_result()
            case instance:
                pass

        self._moved_out = True
        return Ok(instance)

    @override
    async def drop(self, instance: T) -> None:
        if not self._moved_out:
            return

        with contextlib.suppress(StopAsyncIteration):
            await self._gen.asend(instance)

    @override
    def __bool__(self) -> bool:
        return not self._moved_out



from __future__ import annotations

from types import TracebackType
from typing import Any

from greyhorse.maybe import Just, Maybe, Nothing
from greyhorse.utils.invoke import invoke_sync

from ..abc.resources import Lifetime
from ..contexts import CtxCallbacks, SyncContext, SyncContextWithCallbacks
from ..registries import DictRegistry, MutDictRegistry
from .registry import FactoryRegistry, TypeFactory


class Container:
    __slots__ = (
        '_registry',
        '_child_registries',
        '_parent',
        '_resources',
        '_cache',
        '_scoped_factories',
        '_ctx',
    )

    def __init__(
        self,
        registry: FactoryRegistry,
        *child_registries: FactoryRegistry,
        parent: Container | None = None,
        resources: dict[type, Any] | None = None,
    ) -> None:
        self._registry = registry
        self._child_registries = child_registries
        self._parent = parent
        self._resources = MutDictRegistry[type, Any]()
        self._cache = DictRegistry[type, Any]()
        self._scoped_factories: dict[type, TypeFactory] = {}

        if resources:
            for k, v in resources.items():
                self._resources.add(k, v)

        self._ctx = self._create_ctx()

    @property
    def parent(self) -> Container | None:
        return self._parent

    def __repr__(self) -> str:
        return f'Container<{self.lifetime.name}>'

    @property
    def registry(self) -> FactoryRegistry:
        return self._registry

    @property
    def lifetime(self) -> Lifetime:
        return self._registry.lifetime

    @property
    def context(self) -> SyncContext[Container]:
        return self._ctx

    def child_registry(self, lifetime: Lifetime) -> Maybe[FactoryRegistry]:
        for child in self._child_registries:
            if child.lifetime == lifetime:
                return Just(child)
        return Nothing

    def add_resource[T](self, res_type: type[T], resource: T) -> bool:
        return self._resources.add(res_type, resource)

    def remove_resource[T](self, res_type: type[T]) -> bool:
        return self._resources.remove(res_type)

    def __call__(
        self, resources: dict[type, Any] | None = None, lifetime: Lifetime | None = None
    ) -> SyncContext[Container]:
        if lifetime is not None and lifetime.order <= self.lifetime.order:
            return self.context

        if not self._child_registries:
            return self.context

        child = Container(*self._child_registries, parent=self, resources=resources)

        if lifetime is None:
            while child.registry.lifetime.autocreate:
                if not child._child_registries:  # noqa: SLF001
                    break
                child = Container(*child._child_registries, parent=child, resources=resources)  # noqa: SLF001
        else:
            while child.registry.lifetime.order < lifetime.order:
                if not child._child_registries:  # noqa: SLF001
                    raise ValueError(
                        f'Cannot find {lifetime} as a child of current '
                        f'{self._registry.lifetime}'
                    )
                child = Container(*child._child_registries, parent=child, resources=resources)  # noqa: SLF001

        return child.context

    def get[T](self, key: type[T]) -> Maybe[T]:
        return self._get(key)[0]

    def _get[T](self, key: type[T]) -> tuple[Maybe[T], bool]:
        if res := self._cache.get(key):
            return res, True

        if (factory := self._registry.get_factory(key).unwrap_or_none()) and (
            res := invoke_sync(factory.create, key)
        ):
            if factory.scoped:
                self._scoped_factories[key] = factory
            if factory.cache:
                self._cache.add(key, res.unwrap())
            return res, factory.cache

        if not self._parent:
            return Nothing, False

        res, cached = self._parent._get(key)  # noqa: SLF001
        if cached:
            self._cache.add(key, res.unwrap())
        return res, cached

    def _enter(self, _) -> None:  # noqa: ANN001
        if self._parent:
            self._parent.context.__enter__()

        for k, v in self._resources.items():
            self._cache.add(k, v)

    def _exit(
        self,
        _,  # noqa: ANN001
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        while key := next(reversed(self._scoped_factories), None):
            factory = self._scoped_factories[key]
            res = self._resources.get(key).unwrap_or_none()
            invoke_sync(factory.destroy, res)
            del self._scoped_factories[key]

        self._cache.clear()

        if self._parent:
            self._parent.context.__exit__(exc_type, exc_value, traceback)

    def _create_ctx(self) -> SyncContext[Container]:
        return SyncContextWithCallbacks[Container](
            callbacks=CtxCallbacks(on_enter=Just(self._enter), on_exit=Just(self._exit)),
            factory=lambda *_: self,
        )


root: Container = None  # type: ignore


def make_container(
    resources: dict[type, Any] | None = None, lifetime: Lifetime | None = None
) -> Container:
    from ..runtime import Runtime

    global root  # noqa: PLW0602

    runtime = Runtime._instance  # noqa: SLF001
    start_order = 0
    if root is not None:
        start_order = 1
    if runtime is not None:
        start_order = 2

    registries = [FactoryRegistry(s) for s in Lifetime.all() if s.order >= start_order]
    container = None

    if root is not None:
        container = root
    if runtime is not None:
        container = Container(runtime.container.registry, parent=container, resources=resources)

    container = Container(*registries, parent=container, resources=resources)

    if lifetime is None:
        while container.registry.lifetime.autocreate:
            container = Container(
                *container._child_registries,  # noqa: SLF001
                parent=container,
                resources=resources,
            )
    else:
        while container.registry.lifetime.order < lifetime.order:
            container = Container(
                *container._child_registries,  # noqa: SLF001
                parent=container,
                resources=resources,
            )

    return container


root = make_container(lifetime=Lifetime.ROOT())  # type: ignore



from collections.abc import AsyncGenerator, Callable, Generator, Iterable
from functools import cache
from typing import TYPE_CHECKING, Self

from greyhorse.enum import Enum, Struct

type TypeFactoryFn[T] = (
    Callable[[T], T]
    | type[T]
    | T
    | Callable[[], Generator[T, T, None]]
    | Callable[[], AsyncGenerator[T, T]]
)


class Lifetime(Enum):
    ROOT = Struct(name='ROOT', order=0, autocreate=True)
    RUNTIME = Struct(name='RUNTIME', order=1, autocreate=True)
    COMPONENT = Struct(name='COMPONENT', order=2, autocreate=False)
    SESSION = Struct(name='SESSION', order=3, autocreate=True)
    REQUEST = Struct(name='REQUEST', order=4, autocreate=False)
    ACTION = Struct(name='ACTION', order=5, autocreate=True)
    STEP = Struct(name='STEP', order=6, autocreate=False)

    if TYPE_CHECKING:
        name: str
        order: int
        autocreate: str

    @classmethod
    @cache
    def all(cls) -> Iterable[Self]:
        lifetimes = [lifetime_class() for _, lifetime_class in Lifetime.items]
        lifetimes.sort(key=lambda lifetime: lifetime.order)
        return lifetimes



import pickle
from typing import override

from greyhorse.maybe import Just, Maybe, Nothing

from .base import Deserializer, Serializer


class PickleSerializer[T](Serializer[T]):
    def serialize(self, data: T) -> bytes:
        return pickle.dumps(data)


class PickleDeserializer[T](Deserializer[T]):
    @override
    def deserialize(self, data: bytes) -> Maybe[T]:
        try:
            value = pickle.loads(data) if data else None
        except (pickle.UnpicklingError, ValueError, ModuleNotFoundError, MemoryError):
            return Nothing

        return Just(value)



import asyncio
import inspect
import threading
from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable
from dataclasses import dataclass, field
from typing import Any

from greyhorse.enum import Enum, Struct, Tuple, Unit
from greyhorse.error import Error, ErrorCase
from greyhorse.result import Result

from .resources import Lifetime, TypeFactoryFn
from .visitor import Visitor


class ServiceState(Enum):
    Idle = Unit()
    Active = Struct(started=bool)


class ServiceWaiter(Enum):
    Sync = Tuple(threading.Event)
    Async = Tuple(asyncio.Event)


class ServiceError(Error):
    namespace = 'greyhorse.app'

    Factory = ErrorCase(msg='Service factory error: "{details}"', details=str)
    NoSuchResource = ErrorCase(msg='Resource "{name}" is not found', name=str)


type ServiceFactoryFn = Callable[[...], Service | Result[Service, ServiceError]]
type ServiceFactories = dict[type[Service], ServiceFactoryFn]


@dataclass(slots=True, frozen=True)
class ResourceMember:
    resource_type: type
    lifetime: Lifetime
    cache: bool
    method: classmethod
    params: dict[str, type] = field(default_factory=dict)


def _is_resource_member(member: classmethod) -> bool:
    return hasattr(member, '__res_provider__')


class Service(ABC):
    def __init__(self) -> None:
        self._state = ServiceState.Idle
        self._resources: dict[type, TypeFactoryFn] = {}
        self._resource_members: dict[type, ResourceMember] = {}
        self._init_members()

    def accept_visitor(self, visitor: Visitor) -> None:
        visitor.visit_service(self)

    def _init_members(self) -> None:
        resource_members = inspect.getmembers(self, _is_resource_member)

        for _, member in resource_members:
            member = member.__res_provider__
            self._resources[member.resource_type] = member.method
            self._resource_members[member.resource_type] = member

    def inspect(self, callback: Callable[[ResourceMember], Any]) -> None:
        for member in self._resource_members.values():
            callback(member)

    @property
    def state(self) -> ServiceState:
        return self._state

    @property
    @abstractmethod
    def waiter(self) -> ServiceWaiter: ...

    @abstractmethod
    def setup(
        self,
        *args,  # noqa: ANN002
        **kwargs,  # noqa: ANN003
    ) -> Result[ServiceState, ServiceError] | Awaitable[Result[ServiceState, ServiceError]]: ...

    @abstractmethod
    def teardown(
        self,
        *args,  # noqa: ANN002
        **kwargs,  # noqa: ANN003
    ) -> Result[ServiceState, ServiceError] | Awaitable[Result[ServiceState, ServiceError]]: ...

    def _switch_to_idle(self) -> None:
        self._state = ServiceState.Idle

    def _switch_to_active(self, started: bool = False) -> None:
        self._state = ServiceState.Active(started=started)



from __future__ import annotations

import asyncio
import threading
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Awaitable, Callable, Mapping
from contextlib import (
    AbstractAsyncContextManager,
    AbstractContextManager,
    AsyncExitStack,
    ExitStack,
    suppress,
)
from dataclasses import dataclass, field
from types import TracebackType
from typing import Any, override
from uuid import uuid4

from greyhorse.enum import Enum, Struct, Unit
from greyhorse.maybe import Just, Maybe, Nothing
from greyhorse.utils.invoke import (
    get_asyncio_loop,
    invoke_async,
    invoke_sync,
    is_like_async_context_manager,
    is_like_sync_context_manager,
)
from greyhorse.utils.types import TypeWrapper

type FieldFactory[T] = (
    T
    | Callable[[], Awaitable[T] | T]
    | AbstractContextManager[T]
    | AbstractAsyncContextManager[T]
    | Callable[[], AbstractContextManager[T]]
    | Callable[[], AbstractAsyncContextManager[T]]
)


type ContextManagerLike[T] = (
    AbstractContextManager[T]
    | AbstractAsyncContextManager[T]
    | Callable[[], AbstractContextManager[T]]
    | Callable[[], AbstractAsyncContextManager[T]]
)


class ContextState[T](Enum):
    Idle = Unit()
    InUse = Struct(count=int, value=T)
    Applied = Struct(count=int, value=T)
    Cancelled = Struct(count=int, value=T)


@dataclass(slots=True, frozen=True)
class ContextData[T]:
    factory: Callable[[...], T]
    ident: str = field(default_factory=lambda: str(uuid4()))
    field_factories: dict[str, FieldFactory[Any]] = field(default_factory=dict)
    finalizers: list[Callable[[], Awaitable[None] | None]] = field(default_factory=list)


class InvalidContextStateError(RuntimeError):
    pass


class Context:
    __slots__ = ('_state', '_data', '_parent')

    def __init__(
        self,
        factory: Callable[[...], Any],
        fields: Mapping[str, FieldFactory[Any]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
    ) -> None:
        self._state = ContextState[Any].Idle
        self._data = ContextData(
            factory=factory, field_factories=fields or {}, finalizers=finalizers or []
        )
        self._parent: Maybe[Context] = Nothing

    @property
    def ident(self) -> str:
        return self._data.ident

    @property
    def state(self) -> ContextState:
        return self._state

    def __enter__(self) -> object:
        raise NotImplementedError

    def __exit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        raise NotImplementedError

    async def __aenter__(self) -> object:
        raise NotImplementedError

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        raise NotImplementedError

    @property
    def parent(self) -> Maybe[Context]:
        raise NotImplementedError

    def children(self) -> list[Context]:
        raise NotImplementedError


class MutContext(Context, ABC):
    @abstractmethod
    def apply(self) -> Awaitable[None] | None: ...

    @abstractmethod
    def cancel(self) -> Awaitable[None] | None: ...

    @abstractmethod
    def mutable_children(self) -> list[MutContext]: ...


class _ContextStorage:
    __slots__ = ('_storage',)

    def __init__(self) -> None:
        self._storage: dict[type, list[Context]] = defaultdict(list)

    def add(self, kind: type, instance: Context) -> None:
        self._storage[kind].append(instance)

    def remove(self, kind: type, instance: Context) -> None:
        self._storage[kind].remove(instance)

    def get_last(self, kind: type) -> Maybe[Context]:
        if objects := self._storage.get(kind, []):
            return Just(objects[0])
        return Nothing


_context_storage = _ContextStorage()


def current_context_for(kind: type) -> Maybe[Context]:
    return _context_storage.get_last(kind)


def current_scope_id(kind: type[Context] | None = None) -> str:
    if kind is not None and (ctx := _context_storage.get_last(kind)):
        return ctx.unwrap().ident

    if loop := get_asyncio_loop():
        return f'asyncio:{id(asyncio.current_task(loop))}'
    return f'thread:{threading.current_thread().ident}'


class SyncContext[T](Context, TypeWrapper[T], AbstractContextManager):
    __slots__ = ('_sync_stack', '_context_managers', '_children', '_lock')

    def __init__(
        self,
        factory: Callable[[...], T],
        fields: dict[str, FieldFactory[T]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
        sub_contexts: list[ContextManagerLike[T]] | None = None,
    ) -> None:
        fields = fields.copy() if fields else {}
        self._sync_stack = ExitStack()
        self._lock = threading.Lock()
        self._children: list[SyncContext] = []
        self._sub_contexts: list[tuple[ContextManagerLike, str | None]] = []

        if sub_contexts:
            for ctx in sub_contexts:
                if isinstance(ctx, SyncContext):
                    self._children.append(ctx)
                if is_like_sync_context_manager(ctx):
                    self._sub_contexts.append((ctx, None))

        names_to_remove = set()

        for name, value in fields.items():
            if isinstance(value, SyncContext):
                self._children.append(value)
                names_to_remove.add(name)
            if is_like_sync_context_manager(value):
                self._sub_contexts.append((value, name))
                names_to_remove.add(name)
            elif is_like_async_context_manager(value):
                names_to_remove.add(name)

        for name in names_to_remove:
            fields.pop(name)

        super().__init__(factory, fields, finalizers)

    @override
    @property
    def parent(self) -> Maybe[Context]:
        return self._parent

    @override
    def children(self) -> list[Context]:
        return self._children.copy()

    def _switch_to_use(self) -> T:
        self._sync_stack.__enter__()
        kwargs: dict[str, Any] = {}

        for ctx, field in self._sub_contexts:
            if callable(ctx):
                ctx = ctx()
            if (value := self._sync_stack.enter_context(ctx)) and field is not None:
                kwargs[field] = value

        for name, value in self._data.field_factories.items():
            if callable(value):
                value = invoke_sync(value)
            kwargs[name] = value

        instance = self._create(**kwargs)
        self._parent = _context_storage.get_last(type(instance))
        _context_storage.add(type(instance), self)

        return instance

    def _switch_to_idle(self, instance: T, exc_type, exc_value, traceback) -> None:  # noqa: ANN001
        _context_storage.remove(type(instance), self)

        with suppress(Exception):
            self._destroy(instance)

        with suppress(Exception):
            self._sync_stack.__exit__(exc_type, exc_value, traceback)

        for finalizer in self._data.finalizers:
            with suppress(Exception):
                invoke_sync(finalizer)

        self._parent = Nothing

    def _create(self, **kwargs: dict[str, Any]) -> T:
        return self._data.factory(**kwargs)

    def _destroy(self, instance: T) -> None:
        del instance

    def _enter(self, instance: T) -> None:
        pass

    def _exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        pass

    def _nested_enter(self, instance: T) -> None:
        pass

    def _nested_exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        pass

    @override
    def __enter__(self) -> T:
        with self._lock:
            match self._state:
                case ContextState.Idle:
                    value = self._switch_to_use()
                    self._enter(value)
                    self._state = ContextState[type(value)].InUse(count=1, value=value)
                    return value

                case ContextState.InUse(count, value):
                    self._nested_enter(value)
                    self._state = ContextState[type(value)].InUse(count=count + 1, value=value)
                    return value

                case ContextState.Applied(count, value):
                    self._nested_enter(value)
                    self._state = ContextState[type(value)].InUse(count=count + 1, value=value)
                    return value

                case ContextState.Cancelled(count, value):
                    self._nested_enter(value)
                    self._state = ContextState[type(value)].InUse(count=count + 1, value=value)
                    return value

    @override
    def __exit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        with self._lock:
            match self._state:
                case ContextState.Idle:
                    raise InvalidContextStateError('Context exit on idle state')

                case (
                    ContextState.InUse(count, value)
                    | ContextState.Applied(count, value)
                    | ContextState.Cancelled(count, value)
                ):
                    if count > 1:
                        self._nested_exit(value, exc_type, exc_value, traceback)
                        self._state = self._state.__class__(count=count - 1, value=value)
                    else:
                        self._exit(value, exc_type, exc_value, traceback)
                        self._switch_to_idle(value, exc_type, exc_value, traceback)
                        self._state = ContextState[type(value)].Idle


class SyncMutContext[T](SyncContext[T], MutContext):
    __slots__ = ('_mut_children', '_force_rollback', '_auto_apply')

    def __init__(
        self,
        factory: Callable[[...], T],
        fields: dict[str, FieldFactory[T]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
        sub_contexts: list[ContextManagerLike[T]] | None = None,
        force_rollback: bool = False,
        auto_apply: bool = False,
    ) -> None:
        super().__init__(factory, fields, finalizers, sub_contexts)
        self._mut_children = []
        self._force_rollback = force_rollback
        self._auto_apply = auto_apply

        for child in self._children:
            if isinstance(child, SyncMutContext):
                self._mut_children.append(child)

    @override
    def mutable_children(self) -> list[MutContext]:
        return self._mut_children.copy()

    def _apply(self, instance: T) -> None:
        pass

    def _cancel(self, instance: T) -> None:
        pass

    @override
    def apply(self) -> None:
        with self._lock:
            self._do_apply()

    @override
    def cancel(self) -> None:
        with self._lock:
            self._do_cancel()

    @override
    def _exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        if self._force_rollback or exc_type is not None:
            self._do_cancel()
        elif self._auto_apply:
            self._do_apply()

    def _do_apply(self) -> None:
        match self._state:
            case ContextState.Idle:
                raise InvalidContextStateError('MutContext apply on idle state')

            case ContextState.InUse(count, value):
                for child in self._mut_children:
                    child.apply()
                self._apply(value)
                self._state = ContextState[type(value)].Applied(count=count, value=value)

            case ContextState.Applied(_, _):
                pass

            case ContextState.Cancelled(_, _):
                raise InvalidContextStateError('MutContext apply on cancelled state')

    def _do_cancel(self) -> None:
        match self._state:
            case ContextState.Idle:
                raise InvalidContextStateError('MutContext cancel on idle state')

            case ContextState.InUse(count, value):
                for child in self._mut_children:
                    child.cancel()
                self._cancel(value)
                self._state = ContextState[type(value)].Cancelled(count=count, value=value)

            case ContextState.Applied(_, _):
                raise InvalidContextStateError('MutContext cancel on applied state')

            case ContextState.Cancelled(_, _):
                pass


class AsyncContext[T](Context, TypeWrapper[T], AbstractAsyncContextManager):
    __slots__ = (
        '_sync_stack',
        '_async_stack',
        '_sync_sub_contexts',
        '_async_sub_contexts',
        '_sync_children',
        '_async_children',
        '_lock',
    )

    def __init__(
        self,
        factory: Callable[[...], T],
        fields: dict[str, FieldFactory[T]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
        sub_contexts: list[ContextManagerLike[T]] | None = None,
    ) -> None:
        fields = fields.copy() if fields else {}
        self._sync_stack = ExitStack()
        self._async_stack = AsyncExitStack()
        self._lock = asyncio.Lock()
        self._sync_children: list[SyncContext] = []
        self._async_children: list[AsyncContext] = []
        self._sync_sub_contexts: list[tuple[ContextManagerLike[T], str | None]] = []
        self._async_sub_contexts: list[tuple[ContextManagerLike[T], str | None]] = []

        if sub_contexts:
            for ctx in sub_contexts:
                if isinstance(ctx, SyncContext):
                    self._sync_children.append(ctx)
                elif isinstance(ctx, AsyncContext):
                    self._async_children.append(ctx)
                if is_like_async_context_manager(ctx):
                    self._async_sub_contexts.append((ctx, None))
                elif is_like_sync_context_manager(ctx):
                    self._sync_sub_contexts.append((ctx, None))

        names_to_remove = set()

        for name, value in fields.items():
            if isinstance(value, SyncContext):
                self._sync_children.append(value)
                names_to_remove.add(name)
            elif isinstance(value, AsyncContext):
                self._async_children.append(value)
                names_to_remove.add(name)
            if is_like_async_context_manager(value):
                self._async_sub_contexts.append((value, name))
                names_to_remove.add(name)
            elif is_like_sync_context_manager(value):
                self._sync_sub_contexts.append((value, name))
                names_to_remove.add(name)

        for name in names_to_remove:
            fields.pop(name)

        super().__init__(factory, fields, finalizers)

    @override
    @property
    def parent(self) -> Maybe[Context]:
        return self._parent

    @override
    def children(self) -> list[Context]:
        return self._sync_children.copy() + self._async_children.copy()

    async def _switch_to_use(self) -> T:
        self._sync_stack.__enter__()
        await self._async_stack.__aenter__()
        kwargs: dict[str, Any] = {}

        for ctx, field in self._sync_sub_contexts:
            if callable(ctx):
                ctx = ctx()
            if (value := self._sync_stack.enter_context(ctx)) and field is not None:
                kwargs[field] = value

        for ctx, field in self._async_sub_contexts:
            if callable(ctx):
                ctx = ctx()
            if (
                value := await self._async_stack.enter_async_context(ctx)
            ) and field is not None:
                kwargs[field] = value

        for name, value in self._data.field_factories.items():
            if callable(value):
                value = await invoke_async(value)
            kwargs[name] = value

        instance = await self._create(**kwargs)
        self._parent = _context_storage.get_last(type(instance))
        _context_storage.add(type(instance), self)

        return instance

    async def _switch_to_idle(self, instance: T, exc_type, exc_value, traceback) -> None:  # noqa: ANN001
        _context_storage.remove(type(instance), self)

        with suppress(Exception):
            await self._destroy(instance)

        with suppress(Exception):
            await self._async_stack.__aexit__(exc_type, exc_value, traceback)

        with suppress(Exception):
            self._sync_stack.__exit__(exc_type, exc_value, traceback)

        for finalizer in self._data.finalizers:
            with suppress(Exception):
                await invoke_async(finalizer)

        self._parent = Nothing

    async def _create(self, **kwargs: dict[str, Any]) -> T:
        return self._data.factory(**kwargs)

    async def _destroy(self, instance: T) -> None:
        del instance

    async def _enter(self, instance: T) -> None:
        pass

    async def _exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        pass

    async def _nested_enter(self, instance: T) -> None:
        pass

    async def _nested_exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        pass

    @override
    async def __aenter__(self) -> T:
        async with self._lock:
            match self._state:
                case ContextState.Idle:
                    value = await self._switch_to_use()
                    await self._enter(value)
                    self._state = ContextState[type(value)].InUse(count=1, value=value)
                    return value

                case ContextState.InUse(count, value):
                    await self._nested_enter(value)
                    self._state = ContextState[type(value)].InUse(count=count + 1, value=value)
                    return value

                case ContextState.Applied(count, value):
                    await self._nested_enter(value)
                    self._state = ContextState[type(value)].InUse(count=count + 1, value=value)
                    return value

                case ContextState.Cancelled(count, value):
                    await self._nested_enter(value)
                    self._state = ContextState[type(value)].InUse(count=count + 1, value=value)
                    return value

    @override
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        async with self._lock:
            match self._state:
                case ContextState.Idle:
                    raise InvalidContextStateError('Context exit on idle state')

                case (
                    ContextState.InUse(count, value)
                    | ContextState.Applied(count, value)
                    | ContextState.Cancelled(count, value)
                ):
                    if count > 1:
                        await self._nested_exit(value, exc_type, exc_value, traceback)
                        self._state = self._state.__class__(count=count - 1, value=value)
                    else:
                        await self._exit(value, exc_type, exc_value, traceback)
                        await self._switch_to_idle(value, exc_type, exc_value, traceback)
                        self._state = ContextState[type(value)].Idle


class AsyncMutContext[T](AsyncContext[T], MutContext):
    __slots__ = ('_sync_mut_children', '_async_mut_children', '_force_rollback', '_auto_apply')

    def __init__(
        self,
        factory: Callable[[...], T],
        fields: dict[str, FieldFactory[T]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
        sub_contexts: list[ContextManagerLike[T]] | None = None,
        force_rollback: bool = False,
        auto_apply: bool = False,
    ) -> None:
        super().__init__(factory, fields, finalizers, sub_contexts)
        self._sync_mut_children = []
        self._async_mut_children = []
        self._force_rollback = force_rollback
        self._auto_apply = auto_apply

        for child in self._sync_children:
            if isinstance(child, SyncMutContext):
                self._sync_mut_children.append(child)

        for child in self._async_children:
            if isinstance(child, AsyncMutContext):
                self._async_mut_children.append(child)

    @override
    def mutable_children(self) -> list[MutContext]:
        return self._sync_mut_children.copy() + self._async_mut_children.copy()

    async def _apply(self, instance: T) -> None:
        pass

    async def _cancel(self, instance: T) -> None:
        pass

    @override
    async def apply(self) -> None:
        async with self._lock:
            await self._do_apply()

    @override
    async def cancel(self) -> None:
        async with self._lock:
            await self._do_cancel()

    @override
    async def _exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        if self._force_rollback or exc_type is not None:
            await self._do_cancel()
        elif self._auto_apply:
            await self._do_apply()

    async def _do_apply(self) -> None:
        match self._state:
            case ContextState.Idle:
                raise InvalidContextStateError('MutContext apply on idle state')

            case ContextState.InUse(count, value):
                for child in self._sync_mut_children:
                    child.apply()
                for child in self._async_mut_children:
                    await child.apply()
                await self._apply(value)
                self._state = ContextState[type(value)].Applied(count=count, value=value)

            case ContextState.Applied(_, _):
                pass

            case ContextState.Cancelled(_, _):
                raise InvalidContextStateError('MutContext apply on cancelled state')

    async def _do_cancel(self) -> None:
        match self._state:
            case ContextState.Idle:
                raise InvalidContextStateError('MutContext cancel on idle state')

            case ContextState.InUse(count, value):
                for child in self._sync_mut_children:
                    child.cancel()
                for child in self._async_mut_children:
                    await child.cancel()
                await self._cancel(value)
                self._state = ContextState[type(value)].Cancelled(count=count, value=value)

            case ContextState.Applied(_, _):
                raise InvalidContextStateError('MutContext cancel on applied state')

            case ContextState.Cancelled(_, _):
                pass


@dataclass
class CtxCallbacks:
    before_create: Maybe[Callable[[...], Any | Awaitable[Any]]] = Nothing
    after_create: Maybe[Callable[[...], Any | Awaitable[Any]]] = Nothing
    on_destroy: Maybe[Callable[[], Any | Awaitable[Any]]] = Nothing
    on_enter: Maybe[Callable[[...], Any | Awaitable[Any]]] = Nothing
    on_exit: Maybe[Callable[[...], Any | Awaitable[Any]]] = Nothing
    on_nested_enter: Maybe[Callable[[...], Any | Awaitable[Any]]] = Nothing
    on_nested_exit: Maybe[Callable[[...], Any | Awaitable[Any]]] = Nothing


@dataclass
class MutCtxCallbacks(CtxCallbacks):
    on_apply: Maybe[Callable[[...], Any | Awaitable[Any]]] = Nothing
    on_cancel: Maybe[Callable[[...], Any | Awaitable[Any]]] = Nothing


class SyncContextWithCallbacks[T](SyncContext[T]):
    __slots__ = ('_callbacks',)

    def __init__(
        self,
        callbacks: CtxCallbacks,
        factory: Callable[[...], T],
        fields: dict[str, FieldFactory[T]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
        sub_contexts: list[ContextManagerLike[T]] | None = None,
    ) -> None:
        super().__init__(factory, fields, finalizers, sub_contexts)
        self._callbacks = callbacks

    @override
    def _create(self, **kwargs: dict[str, Any]) -> T:
        self._callbacks.before_create.map(lambda f: invoke_sync(f, **kwargs))
        res = self._data.factory(**kwargs)
        self._callbacks.after_create.map(lambda f: invoke_sync(f, res))
        return res

    @override
    def _destroy(self, instance: T) -> None:
        self._callbacks.on_destroy.map(lambda f: invoke_sync(f, instance))  # noqa: F821
        del instance

    @override
    def _enter(self, instance: T) -> None:
        self._callbacks.on_enter.map(lambda f: invoke_sync(f, instance))

    @override
    def _exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        self._callbacks.on_exit.map(
            lambda f: invoke_sync(f, instance, exc_type, exc_value, traceback)
        )

    @override
    def _nested_enter(self, instance: T) -> None:
        self._callbacks.on_nested_enter.map(lambda f: invoke_sync(f, instance))

    @override
    def _nested_exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        self._callbacks.on_nested_exit.map(
            lambda f: invoke_sync(f, instance, exc_type, exc_value, traceback)
        )


class SyncMutContextWithCallbacks[T](SyncMutContext[T]):
    def __init__(
        self,
        callbacks: MutCtxCallbacks,
        factory: Callable[[...], T],
        fields: dict[str, FieldFactory[T]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
        sub_contexts: list[ContextManagerLike[T]] | None = None,
        force_rollback: bool = False,
        auto_apply: bool = False,
    ) -> None:
        super().__init__(factory, fields, finalizers, sub_contexts, force_rollback, auto_apply)
        self._callbacks = callbacks

    @override
    def _create(self, **kwargs: dict[str, Any]) -> T:
        self._callbacks.before_create.map(lambda f: invoke_sync(f, **kwargs))
        res = self._data.factory(**kwargs)
        self._callbacks.after_create.map(lambda f: invoke_sync(res))
        return res

    @override
    def _destroy(self, instance: T) -> None:
        self._callbacks.on_destroy.map(lambda f: invoke_sync(f, instance))  # noqa: F821
        del instance

    @override
    def _enter(self, instance: T) -> None:
        self._callbacks.on_enter.map(lambda f: invoke_sync(f, instance))

    @override
    def _exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        self._callbacks.on_exit.map(
            lambda f: invoke_sync(f, instance, exc_type, exc_value, traceback)
        )

    @override
    def _nested_enter(self, instance: T) -> None:
        self._callbacks.on_nested_enter.map(lambda f: invoke_sync(f, instance))

    @override
    def _nested_exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        self._callbacks.on_nested_exit.map(
            lambda f: invoke_sync(f, instance, exc_type, exc_value, traceback)
        )

    @override
    def _apply(self, instance: T) -> None:
        self._callbacks.on_apply.map(lambda f: invoke_sync(f, instance))

    @override
    def _cancel(self, instance: T) -> None:
        self._callbacks.on_cancel.map(lambda f: invoke_sync(f, instance))


class AsyncContextWithCallbacks[T](AsyncContext[T]):
    __slots__ = ('_callbacks',)

    def __init__(
        self,
        callbacks: CtxCallbacks,
        factory: Callable[[...], T],
        fields: dict[str, FieldFactory[T]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
        sub_contexts: list[ContextManagerLike[T]] | None = None,
    ) -> None:
        super().__init__(factory, fields, finalizers, sub_contexts)
        self._callbacks = callbacks

    @override
    async def _create(self, **kwargs: dict[str, Any]) -> T:
        await self._callbacks.before_create.map_async(lambda f: invoke_async(f, **kwargs))
        res = self._data.factory(**kwargs)
        await self._callbacks.after_create.map_async(lambda f: invoke_async(f, **kwargs))
        return res

    @override
    async def _destroy(self, instance: T) -> None:
        await self._callbacks.on_destroy.map_async(lambda f: invoke_async(f, instance))  # noqa: F821
        del instance

    @override
    async def _enter(self, instance: T) -> None:
        await self._callbacks.on_enter.map_async(lambda f: invoke_async(f, instance))

    @override
    async def _exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        await self._callbacks.on_exit.map_async(
            lambda f: invoke_async(f, instance, exc_type, exc_value, traceback)
        )

    @override
    async def _nested_enter(self, instance: T) -> None:
        await self._callbacks.on_nested_enter.map_async(lambda f: invoke_async(f, instance))

    @override
    async def _nested_exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        await self._callbacks.on_nested_exit.map_async(
            lambda f: invoke_async(f, instance, exc_type, exc_value, traceback)
        )


class AsyncMutContextWithCallbacks[T](AsyncMutContext[T]):
    def __init__(
        self,
        callbacks: MutCtxCallbacks,
        factory: Callable[[...], T],
        fields: dict[str, FieldFactory[T]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
        sub_contexts: list[ContextManagerLike[T]] | None = None,
        force_rollback: bool = False,
        auto_apply: bool = False,
    ) -> None:
        super().__init__(factory, fields, finalizers, sub_contexts, force_rollback, auto_apply)
        self._callbacks = callbacks

    @override
    async def _create(self, **kwargs: dict[str, Any]) -> T:
        await self._callbacks.before_create.map_async(lambda f: invoke_async(f, **kwargs))
        res = self._data.factory(**kwargs)
        await self._callbacks.after_create.map_async(lambda f: invoke_async(f, **kwargs))
        return res

    @override
    async def _destroy(self, instance: T) -> None:
        await self._callbacks.on_destroy.map_async(lambda f: invoke_async(f, instance))  # noqa: F821
        del instance

    @override
    async def _enter(self, instance: T) -> None:
        await self._callbacks.on_enter.map_async(lambda f: invoke_async(f, instance))

    @override
    async def _exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        await self._callbacks.on_exit.map_async(
            lambda f: invoke_async(f, instance, exc_type, exc_value, traceback)
        )

    @override
    async def _nested_enter(self, instance: T) -> None:
        await self._callbacks.on_nested_enter.map_async(lambda f: invoke_async(f, instance))

    @override
    async def _nested_exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        await self._callbacks.on_nested_exit.map_async(
            lambda f: invoke_async(f, instance, exc_type, exc_value, traceback)
        )

    @override
    async def _apply(self, instance: T) -> None:
        await self._callbacks.on_apply.map_async(lambda f: invoke_async(f, instance))

    @override
    async def _cancel(self, instance: T) -> None:
        await self._callbacks.on_cancel.map_async(lambda f: invoke_async(f, instance))


class ContextBuilder[T](TypeWrapper[T]):
    def __init__[**P](
        self,
        factory: Callable[P, T],
        fields: dict[str, FieldFactory[T]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
        sub_contexts: list[ContextManagerLike[T]] | None = None,
        **kwargs: dict[str, Any],
    ) -> None:
        self._factory = factory
        self._fields = fields or {}
        self._finalizers = finalizers or []
        self._sub_contexts = sub_contexts or []
        self._kwargs = kwargs

    def add_param(self, name: str, value: FieldFactory[T]) -> None:
        self._fields[name] = value

    def add_sub_context(self, context: ContextManagerLike[T]) -> None:
        self._sub_contexts.append(context)

    def add_finalizer(self, finalizer: Callable[[], Awaitable[None] | None]) -> None:
        self._finalizers.append(finalizer)

    def build(self) -> T:
        return self.wrapped_type(
            factory=self._factory,
            fields=self._fields,
            finalizers=self._finalizers,
            sub_contexts=self._sub_contexts,
            **self._kwargs,
        )

    def __class_getitem__[C: SyncContext | AsyncContext | SyncMutContext | AsyncMutContext](
        cls, args: tuple[type[C], type[T]]
    ) -> ContextBuilder:
        class_, type_ = args
        return super().__class_getitem__(class_[type_])



import inspect
from dataclasses import dataclass, field
from functools import partial
from typing import Any

from greyhorse.maybe import Just, Maybe, Nothing

from ..abc.resources import Lifetime, TypeFactoryFn
from .factories import TypeFactory


@dataclass(slots=True)
class _ModuleNode:
    default_factory: TypeFactory | None = None
    children: dict[str, '_ModuleNode'] = field(default_factory=dict)
    factories: dict[type, TypeFactory] = field(default_factory=dict)


class FactoryRegistry:
    __slots__ = ('_lifetime', '_root')

    def __init__(self, lifetime: Lifetime) -> None:
        self._lifetime = lifetime
        self._root = _ModuleNode()

    def __repr__(self) -> str:
        return f'FactoryRegistry<{self._lifetime.name}>'

    def __len__(self) -> int:
        nodes = list(self._root.children.values())
        curr_idx = 0

        while curr_idx < len(nodes):
            node = nodes[curr_idx]
            nodes += node.children.values()
            curr_idx += 1

        result = 0

        for node in nodes:
            result += len(node.factories)

        return result

    @property
    def lifetime(self) -> Lifetime:
        return self._lifetime

    @staticmethod
    def _into_factory[T](key: type[T], fn: TypeFactoryFn[T]) -> TypeFactory[T]:
        if callable(fn):
            orig_fn = fn
            while isinstance(orig_fn, partial):
                orig_fn = orig_fn.func

            if inspect.isgeneratorfunction(inspect.unwrap(orig_fn)):
                return TypeFactory[key].from_syncgen(fn)
            if inspect.isasyncgenfunction(inspect.unwrap(orig_fn)):
                return TypeFactory[key].from_asyncgen(fn)
            if inspect.isfunction(orig_fn) or inspect.ismethod(orig_fn):
                return TypeFactory[key].from_fn(fn)
        if inspect.isclass(fn):
            return TypeFactory[key].from_class(fn)
        return TypeFactory[key].from_instance(fn)

    def add_default_factory(
        self, path: str, fn: TypeFactoryFn[Any], cache: bool = False
    ) -> bool:
        factory = self._into_factory(Any, fn)
        factory.cache |= cache
        current = self._root

        if path != '':
            for path_entry in path.split('.'):
                if path_entry not in current.children:
                    current.children[path_entry] = _ModuleNode()
                current = current.children[path_entry]

        if current.default_factory:
            return False

        current.default_factory = factory
        return True

    def add_factory[T](
        self, key: type[T], fn: TypeFactoryFn[T] | None = None, cache: bool = True
    ) -> bool:
        factory = self._into_factory(key, fn or key)
        factory.cache |= cache
        path = key.__module__
        current = self._root

        for path_entry in path.split('.'):
            if path_entry not in current.children:
                current.children[path_entry] = _ModuleNode()
            current = current.children[path_entry]

        if key in current.factories:
            return False

        current.factories[key] = factory
        return True

    def remove_factory[T](self, key: type[T]) -> bool:
        path = key.__module__
        current = self._root

        for path_entry in path.split('.'):
            if path_entry not in current.children:
                return False
            current = current.children[path_entry]

        if key not in current.factories:
            return False

        del current.factories[key]
        return True

    def get_factory[T](self, key: type[T]) -> Maybe[TypeFactory]:
        path = key.__module__
        current = self._root
        nodes: list[_ModuleNode] = []
        wrong_path = False

        for path_entry in path.split('.'):
            if path_entry not in current.children:
                if not current.default_factory:
                    wrong_path = True
                    break
                if current.default_factory.has(key):
                    return Just(current.default_factory)
            else:
                nodes.append(current)
                current = current.children[path_entry]

        if not wrong_path:
            if key in current.factories:
                return Just(current.factories[key])
            if current.default_factory and current.default_factory.has(key):
                return Just(current.default_factory)

        for node in reversed(nodes):
            if node.default_factory and node.default_factory.has(key):
                return Just(node.default_factory)

        return Nothing

    def clear(self) -> None:
        self._root = _ModuleNode()



import sys
import types
from collections.abc import Callable
from typing import cast

import pydantic

from greyhorse.error import Error, ErrorCase
from greyhorse.logging import logger
from greyhorse.result import Ok, Result
from greyhorse.utils.imports import import_path
from greyhorse.utils.injectors import ParamsInjector
from greyhorse.utils.invoke import invoke_sync

from ..schemas.components import ModuleComponentConf, ModuleConf


class ModuleLoadError(Error):
    namespace = 'greyhorse.app.builders.module'

    Disabled = ErrorCase(msg='Module is disabled: "{path}"', path=str)
    AlreadyLoaded = ErrorCase(msg='Module already loaded: "{path}"', path=str)
    WrongImport = ErrorCase(
        msg='Module import error: "{path}", details: "{details}"', path=str, details=str
    )
    Validation = ErrorCase(
        msg='Module validation error: "{path}", details: "{details}"', path=str, details=str
    )
    Unexpected = ErrorCase(
        msg='Module unexpected error: "{path}", details: "{details}"', path=str, details=str
    )
    InvalidConf = ErrorCase(msg='Module conf is invalid: "{path}"', path=str)


class ModuleUnloadError(Error):
    namespace = 'greyhorse.app.builders.module'

    Disabled = ErrorCase(msg='Module is disabled: "{path}"', path=str)
    AlreadyUnloaded = ErrorCase(msg='Module already unloaded: "{path}"', path=str)
    Unexpected = ErrorCase(
        msg='Module unexpected error: "{path}", details: "{details}"', path=str, details=str
    )


class ModuleLoader:
    __slots__ = ('_injector', '_module')

    def __init__(self) -> None:
        self._injector = ParamsInjector()
        self._module: types.ModuleType | None = None

    @property
    def module(self) -> types.ModuleType:
        return self._module

    def load_pass(self, conf: ModuleComponentConf) -> Result[ModuleConf, ModuleLoadError]:
        if conf._conf is not None:  # noqa: SLF001
            return ModuleLoadError.AlreadyLoaded(path=conf.path).to_result()

        if not (res := self._load_module(conf)):
            return res

        conf._conf = res.unwrap()  # noqa: SLF001

        if not conf._conf.enabled:  # noqa: SLF001
            return ModuleLoadError.Disabled(path=conf.path).to_result()

        return Ok(conf._conf)  # noqa: SLF001

    def unload_pass(self, conf: ModuleComponentConf) -> Result[None, ModuleUnloadError]:
        if conf._conf is None:  # noqa: SLF001
            return ModuleUnloadError.AlreadyUnloaded(path=conf.path).to_result()

        if not conf._conf.enabled:  # noqa: SLF001
            return ModuleUnloadError.Disabled(path=conf.path).to_result()

        return self._unload_module(conf)

    @staticmethod
    def _get_module_package(conf: ModuleComponentConf) -> str:
        if not conf.path.startswith('.'):
            return conf.path

        path = conf.path
        dots_count = 0

        for c in path[1:]:
            if c == '.':
                dots_count += 1
            else:
                path = path[dots_count + 1 :]
                break

        init_path = conf._init_path  # noqa: SLF001
        dots_count = min(dots_count, len(init_path))
        init_path = init_path[0 : len(init_path) - dots_count]
        return '.'.join([*init_path, path])

    def _load_module(self, conf: ModuleComponentConf) -> Result[ModuleConf, ModuleLoadError]:
        module_path = self._get_module_package(conf)
        logger.info('ModuleLoader: Try to load module "{path}"'.format(path=module_path))

        try:
            func = cast(Callable[[...], ModuleConf], import_path(f'{module_path}:__init__'))

        except (ImportError, AttributeError) as e:
            error = ModuleLoadError.WrongImport(path=module_path, details=str(e))
            logger.error(error.message)
            return error.to_result()

        injected_args = self._injector(func, values=conf.args)

        try:
            res = invoke_sync(func, *injected_args.args, **injected_args.kwargs)

        except pydantic.ValidationError as e:
            error = ModuleLoadError.Validation(path=module_path, details=str(e))
            logger.error(error.message)
            return error.to_result()
        except Exception as e:
            error = ModuleLoadError.Unexpected(path=module_path, details=str(e))
            logger.error(error.message)
            return error.to_result()

        if not isinstance(res, ModuleConf):
            error = ModuleLoadError.InvalidConf(path=module_path)
            logger.error(error.message)
            return error.to_result()

        self._module = sys.modules[sys.modules[module_path].__package__]

        logger.info(
            'ModuleLoader: Module "{path}" loaded successfully'.format(path=module_path)
        )
        return Ok(res)

    def _unload_module(self, conf: ModuleComponentConf) -> Result[None, ModuleUnloadError]:
        module_path = self._get_module_package(conf)
        logger.info('ModuleLoader: Try to unload module "{path}"'.format(path=module_path))

        try:
            func = cast(Callable[[...], None], import_path(f'{module_path}:__fini__'))

        except (ImportError, AttributeError):
            pass
        else:
            injected_args = self._injector(func, types={ModuleConf: conf._conf})  # noqa: SLF001

            try:
                invoke_sync(func, *injected_args.args, **injected_args.kwargs)

            except Exception as e:
                error = ModuleUnloadError.Unexpected(path=module_path, details=str(e))
                logger.error(error.message)
                return error.to_result()

        conf, conf._conf = conf._conf, None  # noqa: SLF001
        del conf
        del sys.modules[module_path]

        logger.info(
            'ModuleLoader: Module "{path}" unloaded successfully'.format(path=module_path)
        )
        return Ok()



from collections.abc import AsyncGenerator, Generator
from decimal import Decimal

from greyhorse.app.resources.container import make_container, root
from greyhorse.app.resources.injection import inject


def sync_ctx_manager() -> Generator[str, str, None]:
    res = '123'
    yield res
    assert res == '123'


async def async_ctx_manager() -> AsyncGenerator[Decimal, Decimal]:
    res = Decimal(789)
    yield res
    assert res == Decimal(789)


def test_container() -> None:
    container = make_container()

    assert container.registry.add_factory(int, lambda _: 123)
    assert container.registry.add_factory(str, sync_ctx_manager)
    assert container.registry.add_factory(Decimal, async_ctx_manager)

    assert container.get(int).unwrap() == 123

    with container.context:
        with container() as c1:
            assert c1.registry.add_factory(int, lambda _: 456)
            assert c1.get(int).unwrap() == 456
            assert c1.get(str).unwrap() == '123'
            assert Decimal(789) == c1.get(Decimal).unwrap()

        assert container.get(int).unwrap() == 123


class A:
    pass


class B:
    @inject
    def __init__(self, a: A) -> None:
        self.a = a


class C:
    @inject
    def __init__(self, conf: str) -> None:
        self._conf = conf

    def get_data(self) -> str:
        return self._conf


@inject
def make_c(deps: C | None) -> C | None:
    return deps


class SomeClass:
    @inject
    def __init__(self, a: A, b: B, c: C | None) -> None:
        self.a = a
        self.b = b
        self.c = c
        c1 = make_c()
        assert c is c1


def test_injection() -> None:
    conf_str = 'test'
    a = A()

    def factory(t: type) -> object:
        match t.__name__:
            case 'A':
                return a
            case 'B':
                return B()
            case 'C':
                return C()
            case _:
                return None

    assert root.registry.add_default_factory('', factory, cache=True)
    assert root.registry.add_factory(str, conf_str)

    instance = SomeClass()

    assert isinstance(instance.a, A)
    assert isinstance(instance.b, B)
    assert isinstance(instance.c, C)

    assert instance.a is a
    assert instance.b.a is a
    assert instance.c.get_data() is conf_str
